import warnings
from typing import Literal

import jax.numpy as jnp
import numpy as np
import pytest
from jaxtyping import Array, Float, install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    import cryojax.ndimage as im
    import cryojax.simulator as cxs
    from cryojax.constants import (
        LobatoScatteringFactorParameters,
        PengScatteringFactorParameters,
        check_atomic_numbers_supported,
    )
    from cryojax.io import read_array_from_mrc, read_atoms_from_pdb
    from cryojax.ndimage import make_coordinate_grid

try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


@pytest.fixture
def pdb_info(sample_pdb_path):
    return read_atoms_from_pdb(sample_pdb_path, center=True, loads_properties=True)


@pytest.fixture
def toy_gaussian_cloud():
    atom_positions = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )
    num_atoms = atom_positions.shape[0]
    ff_a = jnp.array(
        num_atoms
        * [
            [1.0, 0.5],
        ]
    )

    ff_b = jnp.array(
        num_atoms
        * [
            [0.3, 0.2],
        ]
    )

    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.05
    return (atom_positions, ff_a, ff_b, n_voxels_per_side, voxel_size)


#
# Test volume tabulations
#
@pytest.mark.parametrize("tabulation", ("peng", "lobato"))
def test_load_atom_volume(tabulation, sample_pdb_path: str):
    import pathlib

    import mmdf

    atom_volume = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.IndependentAtomVolume,
        tabulation=tabulation,
    )
    assert isinstance(atom_volume, cxs.IndependentAtomVolume)
    if tabulation == "peng":
        atom_volume = cxs.load_tabulated_volume(
            sample_pdb_path,
            output_type=cxs.GaussianMixtureVolume,
            tabulation=tabulation,
        )
        assert isinstance(atom_volume, cxs.GaussianMixtureVolume)
    else:
        with pytest.raises(ValueError):
            atom_volume = cxs.load_tabulated_volume(
                sample_pdb_path,
                output_type=cxs.GaussianMixtureVolume,
                tabulation=tabulation,
            )
    atom_data = mmdf.read(pathlib.Path(sample_pdb_path))
    atom_volume = cxs.load_tabulated_volume(
        atom_data,
        output_type=cxs.IndependentAtomVolume,
        tabulation=tabulation,
    )
    assert isinstance(atom_volume, cxs.IndependentAtomVolume)


def test_scattering_factor_parameters_correct(
    peng_parameters_path, lobato_parameters_path
):
    from cryojax.constants._scattering_factor_parameters import _SUPPORTED_ATOMIC_NUMBERS

    atomic_numbers = np.asarray(_SUPPORTED_ATOMIC_NUMBERS)

    # Test Peng
    params = PengScatteringFactorParameters(atomic_numbers)
    a1, b1 = (params.a, params.b)

    peng_table = np.load(peng_parameters_path)
    a2, b2 = peng_table[:, atomic_numbers, :]

    np.testing.assert_equal(a1, a2)
    np.testing.assert_equal(b1, b2)

    # Test lobato
    params = LobatoScatteringFactorParameters(atomic_numbers)
    a1, b1 = (params.a, params.b)

    lobato_table = np.load(lobato_parameters_path)
    a2, b2 = lobato_table[:, atomic_numbers, :]

    np.testing.assert_equal(a1, a2)
    np.testing.assert_equal(b1, b2)


def test_compare_hydrogen_scattering_factor():
    shape, pixel_size = (32, 32), 1.0
    frequencies = im.make_frequency_grid(shape, pixel_size)
    hydrogen_id = np.asarray([1], dtype=int)
    p, l = (
        PengScatteringFactorParameters(hydrogen_id),
        LobatoScatteringFactorParameters(hydrogen_id),
    )
    p_fac = cxs.PengScatteringFactor(p.a[0], p.b[0])
    l_fac = cxs.LobatoScatteringFactor(l.a[0], l.b[0])
    p_arr, l_arr = p_fac(frequencies), l_fac(frequencies)
    np.testing.assert_allclose(p_arr, l_arr, atol=1e-3)


@pytest.mark.parametrize("tabulation", ("peng", "lobato"))
def test_invalid_atomic_numbers(tabulation):
    # Make sure has nan when expected
    bad_nan = np.asarray([2, 6])
    params_nan = _make_scattering_parameters(bad_nan, tabulation)
    assert np.any(np.isnan(params_nan.a))
    assert np.any(np.isnan(params_nan.b))
    # Make sure throws out of bounds error when expected
    bad_oob = np.asarray([1, 31])
    with pytest.raises(IndexError):
        _make_scattering_parameters(bad_oob, tabulation)
    # Make sure error checks work
    with pytest.raises(ValueError):
        check_atomic_numbers_supported(bad_nan)
    with pytest.raises(ValueError):
        check_atomic_numbers_supported(bad_oob)


#
# Test volume rendering
#
def test_render_voxels(sample_pdb_path):
    atom_volume = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.GaussianMixtureVolume,
    )
    render_fn = cxs.AutoVolumeRenderFn((16, 16, 16), voxel_size=4.0)
    for cls in [
        cxs.FourierVoxelGridVolume,
        cxs.FourierVoxelSplineVolume,
        cxs.RealVoxelGridVolume,
    ]:
        assert (
            type(cxs.render_voxel_volume(atom_volume, render_fn, output_type=cls)) == cls
        )


def test_voxel_volume_loaders():
    real_voxel_grid = jnp.zeros((10, 10, 10), dtype=float)
    fourier_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)
    real_volume = cxs.RealVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)

    assert isinstance(
        fourier_volume.frequency_slice_in_pixels,
        Float[Array, "1 _ _ 3"],  # type: ignore
    )
    assert isinstance(real_volume.coordinate_grid_in_pixels, Float[Array, "_ _ _ 3"])  # type: ignore


@pytest.mark.parametrize("pad_scale", (1, 1.1))
def test_sinc_correction(sample_mrc_path, pad_scale):
    real_voxel_grid = read_array_from_mrc(sample_mrc_path)
    _ = cxs.FourierVoxelGridVolume.from_real_voxel_grid(
        real_voxel_grid, sinc_correction=True, pad_scale=pad_scale
    )


def test_fourier_vs_real_agreement(sample_pdb_path):
    """
    Integration test ensuring that the VoxelGrid classes
    produce comparable electron densities when loaded from PDB.
    """
    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.5

    atom_volume = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.GaussianMixtureVolume,
        selection_string="not element H",
    )
    render_fn = cxs.AutoVolumeRenderFn(
        n_voxels_per_side,
        voxel_size,
    )
    fourier_volume = cxs.render_voxel_volume(
        atom_volume, render_fn, output_type=cxs.FourierVoxelGridVolume
    )
    real_volume = cxs.render_voxel_volume(
        atom_volume, render_fn, output_type=cxs.RealVoxelGridVolume
    )
    real_voxel_grid = im.ifftn(jnp.fft.ifftshift(fourier_volume.fourier_voxel_grid)).real

    np.testing.assert_allclose(real_voxel_grid, real_volume.real_voxel_grid, atol=1e-12)


def test_downsampled_voxel_volume_agreement(sample_pdb_path):
    """Integration test ensuring that rasterized voxel grids roughly
    agree with downsampled versions.
    """
    # Parameters for rasterization
    shape = (128, 128, 128)
    voxel_size = 0.25
    # Downsampling parameters
    downsampling_factor = 2
    downsampled_shape = (
        int(shape[0] / downsampling_factor),
        int(shape[1] / downsampling_factor),
        int(shape[2] / downsampling_factor),
    )
    downsampled_voxel_size = voxel_size * downsampling_factor
    # Load the PDB file
    atom_positions, atom_types = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="not element H",
    )
    # Load atomistic volume
    atom_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        parameters=PengScatteringFactorParameters(atom_types),
    )
    # Build the grids
    lowres_render_fn = cxs.GaussianMixtureRenderFn(
        downsampled_shape, downsampled_voxel_size
    )
    low_resolution_volume_grid = lowres_render_fn(atom_volume)
    highres_render_fn = cxs.GaussianMixtureRenderFn(shape, voxel_size)
    high_resolution_volume_grid = highres_render_fn(atom_volume)
    downsampled_volume_grid = im.fourier_crop_downsample(
        high_resolution_volume_grid, downsampling_factor
    )

    assert low_resolution_volume_grid.shape == downsampled_volume_grid.shape


def test_render_options(pdb_info):
    width, voxel_size, shape = (1.0, 1.0, (31, 32, 33))
    atom_positions, _, _ = pdb_info
    volumes, render_fns = [], []
    gaussian_volume, gaussian_render_fn = (
        cxs.GaussianMixtureVolume(
            atom_positions,
            amplitudes=1.0,
            variances=width**2,
        ),
        cxs.GaussianMixtureRenderFn(shape, voxel_size),
    )
    volumes.append(gaussian_volume)
    render_fns.append(gaussian_render_fn)
    if jnufft is not None:
        volumes.append(
            cxs.IndependentAtomVolume(
                positions=atom_positions,
                scattering_factors=im.FourierGaussian(
                    amplitude=1.0, b_factor=width**2 * (8 * np.pi**2)
                ),
            )
        )
        render_fns.append(cxs.FFTAtomRenderFn(shape, voxel_size, eps=1e-16))
    for volume, render_fn in zip(volumes, render_fns):
        real_voxel_grid = render_fn(volume, outputs_real_space=True)
        assert real_voxel_grid.shape == shape
        assert not jnp.iscomplexobj(real_voxel_grid)
        fftn_voxel_grid = render_fn(
            volume,
            outputs_real_space=False,
            outputs_rfft=False,
        )
        assert fftn_voxel_grid.shape == shape
        assert jnp.iscomplexobj(fftn_voxel_grid)
        rfftn_voxel_grid = render_fn(
            volume,
            outputs_real_space=False,
            outputs_rfft=True,
        )
        assert rfftn_voxel_grid.shape == (*shape[0:2], shape[2] // 2 + 1)
        assert jnp.iscomplexobj(rfftn_voxel_grid)


@pytest.mark.parametrize(
    "width, voxel_size, shape",
    ((1.0, 0.5, (64, 64, 64)), (1.0, 0.5, (63, 63, 63))),
)
def test_fft_atom_render(pdb_info, width, voxel_size, shape):
    if jnufft is not None:
        atom_positions, _, _ = pdb_info
        gaussian_volume = cxs.GaussianMixtureVolume(
            atom_positions,
            amplitudes=1.0,
            variances=width**2,
        )
        atom_volume = cxs.IndependentAtomVolume(
            positions=atom_positions,
            scattering_factors=im.FourierGaussian(
                amplitude=1.0, b_factor=width**2 * (8 * np.pi**2)
            ),
        )
        gaussian_render_fn = cxs.GaussianMixtureRenderFn(shape, voxel_size)
        fft_render_fn = cxs.FFTAtomRenderFn(shape, voxel_size, eps=1e-16)
        voxels_by_gaussians = gaussian_render_fn(gaussian_volume)
        voxels_by_fft = fft_render_fn(atom_volume)

        np.testing.assert_allclose(voxels_by_gaussians, voxels_by_fft, atol=1e-8)
    else:
        warnings.warn(
            "Could not test rendering method `FFTAtomRenderFn`, "
            "most likely because `jax_finufft` is not installed. "
            f"Error traceback is:\n{JAX_FINUFFT_IMPORT_ERROR}"
        )


#
# TODO: organize
#
@pytest.mark.parametrize("shape", ((128, 127, 126),))
def test_compute_rectangular_voxel_grid(sample_pdb_path, shape):
    voxel_size = 0.5

    # Load the PDB file
    atom_positions, atom_types = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="not element H",
    )
    # Load atomistic volume
    atom_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        parameters=PengScatteringFactorParameters(atom_types),
    )
    # Build the grid
    render_fn = cxs.GaussianMixtureRenderFn(shape, voxel_size)
    voxels = render_fn(atom_volume)
    assert voxels.shape == shape


@pytest.mark.parametrize(
    "batch_size, n_batches",
    ((1, 1), (2, 1), (3, 1), (1, 2), (1, 3), (2, 2)),
)
def test_z_plane_batched_vs_non_batched_loop_agreement(
    sample_pdb_path, batch_size, n_batches
):
    shape = (128, 128, 128)
    voxel_size = 0.5

    # Load the PDB file
    atom_positions, atom_types = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        loads_b_factors=False,
        selection_string="not element H",
    )
    # Load atomistic volume
    atom_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        parameters=PengScatteringFactorParameters(atom_types),
    )
    # Build the grid
    render_fn = cxs.GaussianMixtureRenderFn(shape, voxel_size)
    voxels = render_fn(atom_volume)
    batched_render_fn = cxs.GaussianMixtureRenderFn(
        shape,
        voxel_size,
        batch_options=dict(batch_size=batch_size, n_batches=n_batches),
    )
    voxels_with_batching = batched_render_fn(atom_volume)
    np.testing.assert_allclose(voxels, voxels_with_batching)


class TestIntegrateGMMToPixels:
    @pytest.mark.parametrize("largest_atom", range(0, 3))
    def test_maxima_are_in_right_positions(self, toy_gaussian_cloud, largest_atom):
        """
        Test that the maxima of the volume are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        n_pixels_per_side = n_voxels_per_side[:2]
        ff_a = ff_a.at[largest_atom].add(1.0)
        coordinate_grid = make_coordinate_grid(n_pixels_per_side, voxel_size)

        # Build the volume
        atomic_volume = cxs.GaussianMixtureVolume(
            atom_positions, ff_a, ff_b / (8.0 * jnp.pi**2)
        )
        image_config = cxs.BasicImageConfig(
            shape=n_pixels_per_side,
            pixel_size=voxel_size,
            voltage_in_kilovolts=300.0,
        )
        # Build the volume integrators
        integrator = cxs.GaussianMixtureProjection()
        # Compute projections
        projection = integrator.integrate(atomic_volume, image_config)
        projection = im.irfftn(projection)

        # Find the maximum
        maximum_index = jnp.argmax(projection)
        maximum_position = coordinate_grid.reshape(-1, 2)[maximum_index]

        # Check that the maximum is in the correct position
        assert jnp.allclose(maximum_position, atom_positions[largest_atom][:2])

    def test_integral_is_correct(self, toy_gaussian_cloud):
        """
        Test that the maxima of the volume are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        n_pixels_per_side = n_voxels_per_side[:2]
        # Build the volume
        atomic_volume = cxs.GaussianMixtureVolume(
            atom_positions, ff_a, ff_b / (8.0 * jnp.pi**2)
        )
        image_config = cxs.BasicImageConfig(
            shape=n_pixels_per_side,
            pixel_size=voxel_size,
            voltage_in_kilovolts=300.0,
        )
        # Build the volume integrators
        integrator = cxs.GaussianMixtureProjection()
        # Compute projections
        projection = integrator.integrate(atomic_volume, image_config)
        projection = im.irfftn(projection)

        integral = jnp.sum(projection) * voxel_size**2
        assert jnp.isclose(integral, jnp.sum(ff_a))


class TestRenderGMMToVoxels:
    @pytest.mark.parametrize("largest_atom", range(0, 3))
    def test_maxima_are_in_right_positions(self, toy_gaussian_cloud, largest_atom):
        """
        Test that the maxima of the volume are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        ff_a = ff_a.at[largest_atom].add(1.0)

        # Build the volume
        gmm_volume = cxs.GaussianMixtureVolume(
            atom_positions, ff_a, ff_b / (8 * jnp.pi**2)
        )
        render_fn = cxs.GaussianMixtureRenderFn(n_voxels_per_side, voxel_size)
        real_voxel_grid = render_fn(gmm_volume)
        coordinate_grid = make_coordinate_grid(n_voxels_per_side, voxel_size)

        # Find the maximum
        maximum_index = jnp.argmax(real_voxel_grid)
        maximum_position = coordinate_grid.reshape(-1, 3)[maximum_index]

        # Check that the maximum is in the correct position
        assert jnp.allclose(maximum_position, atom_positions[largest_atom])

    def test_integral_is_correct(self, toy_gaussian_cloud):
        """
        Test that the maxima of the volume are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        # Build the volume
        gmm_volume = cxs.GaussianMixtureVolume(
            atom_positions, ff_a, ff_b / (8 * jnp.pi**2)
        )
        render_fn = cxs.GaussianMixtureRenderFn(n_voxels_per_side, voxel_size)
        real_voxel_grid = render_fn(gmm_volume)

        integral = jnp.sum(real_voxel_grid) * voxel_size**3
        assert jnp.isclose(integral, jnp.sum(ff_a))


def test_gmm_shape():
    n_atoms, n_gaussians = 10, 2
    pos = np.zeros((n_atoms, 3))
    make_gmm = lambda amp, var: cxs.GaussianMixtureVolume(pos, amp, var)
    gmm = make_gmm(1.0, 1.0)
    assert gmm.variances.shape == gmm.amplitudes.shape == (n_atoms, 1)
    gmm = make_gmm(np.ones((n_atoms,)), np.ones((n_atoms,)))
    assert gmm.variances.shape == gmm.amplitudes.shape == (n_atoms, 1)
    gmm = make_gmm(np.ones((n_atoms, n_gaussians)), np.ones((n_atoms, n_gaussians)))
    assert gmm.variances.shape == gmm.amplitudes.shape == (n_atoms, n_gaussians)
    gmm1, gmm2 = make_gmm(1.0, np.ones((n_atoms,))), make_gmm(np.ones((n_atoms,)), 1.0)
    assert (
        gmm1.variances.shape
        == gmm1.amplitudes.shape
        == gmm2.variances.shape
        == gmm2.amplitudes.shape
        == (n_atoms, 1)
    )
    gmm1, gmm2 = (
        make_gmm(1.0, np.ones((n_atoms, n_gaussians))),
        make_gmm(np.ones((n_atoms, n_gaussians)), 1.0),
    )
    assert (
        gmm1.variances.shape
        == gmm1.amplitudes.shape
        == gmm2.variances.shape
        == gmm2.amplitudes.shape
        == (n_atoms, n_gaussians)
    )
    gmm1, gmm2 = (
        make_gmm(np.asarray(1.0), np.ones((n_atoms, n_gaussians))),
        make_gmm(np.ones((n_atoms, n_gaussians)), np.asarray(1.0)),
    )
    assert (
        gmm1.variances.shape
        == gmm1.amplitudes.shape
        == gmm2.variances.shape
        == gmm2.amplitudes.shape
        == (n_atoms, n_gaussians)
    )
    gmm1, gmm2 = (
        make_gmm(np.ones((n_atoms,)), np.ones((n_atoms, n_gaussians))),
        make_gmm(np.ones((n_atoms, n_gaussians)), np.ones((n_atoms,))),
    )
    assert (
        gmm1.variances.shape
        == gmm1.amplitudes.shape
        == gmm2.variances.shape
        == gmm2.amplitudes.shape
        == (n_atoms, n_gaussians)
    )


def _make_scattering_parameters(
    atomic_numbers: np.ndarray, tabulation: Literal["peng", "lobato"]
):
    return (
        PengScatteringFactorParameters(atomic_numbers)
        if tabulation == "peng"
        else LobatoScatteringFactorParameters(atomic_numbers)
    )



from ._conventions import (
    b_factor_to_variance as b_factor_to_variance,
    variance_to_b_factor as variance_to_b_factor,
)
from ._parkhurst2024_solvent_power import (
    PARKHURST2024_POWER_CONSTANTS as PARKHURST2024_POWER_CONSTANTS,
)
from ._physical_constants import (
    interaction_constant_from_kilovolts as interaction_constant_from_kilovolts,
    lorentz_factor_from_kilovolts as lorentz_factor_from_kilovolts,
    wavelength_from_kilovolts as wavelength_from_kilovolts,
)
from ._scattering_factor_parameters import (
    LobatoScatteringFactorParameters as LobatoScatteringFactorParameters,
    PengScatteringFactorParameters as PengScatteringFactorParameters,
    check_atomic_numbers_supported as check_atomic_numbers_supported,
)



from typing import TypeVar

import jax
import numpy as np
from jaxtyping import Int, PyTree


T = TypeVar("T")


def split_atoms_by_element(
    atomic_numbers: Int[np.ndarray, " _"],
    atom_pytree: PyTree[np.ndarray, "T"],
    atom_axis: int = 0,
) -> tuple[PyTree[tuple[np.ndarray, ...], "T"], Int[np.ndarray, " _"]]:
    """Given atomic numbers, split a pytree of numpy arrays
    representing atom properties into a tuple where each element
    is the pytree of tuples where each element is the array for
    a given atomic number.

    **Arguments:**

    - `atomic_numbers`:
        Atomic numbers as a numpy arrays.
    - `atom_pytree`:
        A pytree of numpy arrays,
    - `atom_axis`:
        The axis representing the atom index. Leading axes
        are assumed to be batched axes.

    **Returns:**

    A pytree with tree structure matching `atom_pytree`,
    where arrays have been replaced with tuples of arrays.
    """
    unique_atomic_numbers = np.unique(atomic_numbers)
    split_pytree = jax.tree.map(
        lambda x: tuple(
            np.take(x, np.where(atomic_numbers == id)[0], axis=atom_axis)
            for id in unique_atomic_numbers
        ),
        atom_pytree,
    )
    return split_pytree, unique_atomic_numbers



"""
These versions of the scipy function map_coordinates is modified from Louis Desdoigts's
version: https://github.com/LouisDesdoigts/jax/blob/cubic-spline-updated/jax/_src/scipy/ndimage.py.

This code was developed for the project [`dLux`](https://louisdesdoigts.github.io/dLux/).
"""

import functools
import itertools
import operator
from collections.abc import Sequence

import jax.numpy as jnp
import lineax as lx
from jax import lax, vmap
from jaxtyping import Array, ArrayLike


def map_coordinates(
    input: Array,
    coordinates: Sequence[Array],
    order: int = 1,
    mode: str = "fill",
    cval: float | complex = 0.0,
):
    """
    Similar to `scipy.map_coordinates`, but diverges from the API.

    Adapted from Louis Desdoigts's [version of `jax.scipy.map_coordinates`](https://github.com/LouisDesdoigts/jax/blob/cubic-spline-updated/jax/_src/scipy/ndimage.py),
    which was developed for the project [`dLux`](https://louisdesdoigts.github.io/dLux/).

    **Arguments:**

    - `mode` :
        Uses built-in [JAX out-of-bounds indexing](https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html)
        to determine how to extrapolate beyond boundaries.
    """  # noqa: E501
    if order in [0, 1]:
        return _map_coordinates_nn_or_linear(input, coordinates, order, mode, cval)
    elif order == 3:
        coefficients = compute_spline_coefficients(input)
        return _map_coordinates_with_cubic_spline(coefficients, coordinates, mode, cval)
    else:
        raise NotImplementedError(f"map_coordinates does not support order={order}.")


def map_coordinates_spline(
    coefficients: Array,
    coordinates: Sequence[Array],
    mode: str = "fill",
    cval: float | complex = 0.0,
):
    """
    Similar to `scipy.map_coordinates`, but takes coefficients computed from
    [`cryojax.ndimage.compute_spline_coefficients`][] as input.
    """
    return _map_coordinates_with_cubic_spline(coefficients, coordinates, mode, cval)


def compute_spline_coefficients(data: Array) -> Array:
    """Solve for the spline coefficients of an input array."""
    ndim = data.ndim
    for i in range(ndim):
        axis = ndim - i - 1
        A = _build_operator(data.shape[axis] - 2, dtype=data.dtype)
        fn = lambda x: _solve_coefficients(x, A)
        for j in range(ndim - 2, -1, -1):
            ax = int(j >= axis)
            fn = vmap(fn, ax, ax)
        data = fn(data)
    return data


def _map_coordinates_nn_or_linear(
    input: ArrayLike,
    coordinates: Sequence[ArrayLike],
    order: int,
    mode: str,
    cval: float | complex,
) -> Array:
    input_arr = jnp.asarray(input)
    coordinate_arrs = [jnp.asarray(c) for c in coordinates]

    if len(coordinates) != input_arr.ndim:
        raise ValueError(
            f"Coordinates must be a sequence of length {input_arr.ndim}, "
            f"but found that it was equal to {len(coordinates)}."
        )

    if order == 0:
        interp_fun = _nearest_indices_and_weights
    elif order == 1:
        interp_fun = _linear_indices_and_weights
    else:
        raise NotImplementedError(
            "_map_coordinates_nn_or_linear requires order = 0 or 1."
        )

    interpolations_1d = []
    for coordinate in coordinate_arrs:
        interp_nodes = interp_fun(coordinate)
        interpolations_1d.append(interp_nodes)
    outputs = []
    for items in itertools.product(*interpolations_1d):
        indices, weights = _unzip2(items)
        contribution = input_arr.at[indices].get(mode=mode, fill_value=cval)
        interpolated = _nonempty_prod(weights) * contribution
        outputs.append(interpolated)
    result = _nonempty_sum(outputs)
    if jnp.issubdtype(input_arr.dtype, jnp.integer):
        result = _round_half_away_from_zero(result)
    return result.astype(input_arr.dtype)


def _map_coordinates_with_cubic_spline(
    coefficients: ArrayLike,
    coordinates: Sequence[ArrayLike],
    mode: str,
    cval: complex | float,
) -> Array:
    coefficients = jnp.asarray(coefficients)
    if len(coordinates) != coefficients.ndim:
        raise ValueError(
            "coordinates must be a sequence of length coefficients.ndim, but "
            f"{len(coordinates)} != {coefficients.ndim}"
        )
    # Stack coordinates along one axis
    coords = jnp.stack([jnp.asarray(c) for c in coordinates], axis=0)
    # vmap spline evaluate over coordinates and return
    fn = lambda coord: _spline_point(coefficients, coord, mode, cval)
    return vmap(fn)(coords.reshape(coefficients.ndim, -1).T).reshape(coords.shape[1:])


#
# Nearest neighbor and linear interpolation utilities
#
def _nonempty_prod(arrs: Sequence[Array]) -> Array:
    return functools.reduce(operator.mul, arrs)


def _nonempty_sum(arrs: Sequence[Array]) -> Array:
    return functools.reduce(operator.add, arrs)


def _round_half_away_from_zero(a: Array) -> Array:
    return a if jnp.issubdtype(a.dtype, jnp.integer) else lax.round(a)


def _nearest_indices_and_weights(
    coordinate: Array,
) -> list[tuple[Array, ArrayLike]]:
    index = _round_half_away_from_zero(coordinate).astype(jnp.int32)
    weight = coordinate.dtype.type(1)
    return [(index, weight)]


def _linear_indices_and_weights(
    coordinate: Array,
) -> list[tuple[Array, ArrayLike]]:
    lower = jnp.floor(coordinate)
    upper_weight = coordinate - lower
    lower_weight = 1 - upper_weight
    index = lower.astype(jnp.int32)
    return [(index, lower_weight), (index + 1, upper_weight)]


#
# Spline interpolation utilities
#
def _build_operator(
    n: int, dtype: jnp.dtype | None = None, diag_value: float = 4.0
) -> lx.TridiagonalLinearOperator:
    diagonal = jnp.full((n,), diag_value, dtype=dtype)
    lower_diagonal = jnp.full((n - 1,), 1.0, dtype=dtype)
    upper_diagonal = jnp.full((n - 1,), 1.0, dtype=dtype)
    return lx.TridiagonalLinearOperator(diagonal, lower_diagonal, upper_diagonal)


def _construct_vector(data: Array, c2: Array, cnp2: Array) -> Array:
    yvec = data[1:-1]
    first = data[1] - c2
    last = data[-2] - cnp2
    yvec = yvec.at[0].set(first)
    yvec = yvec.at[-1].set(last)
    return yvec


def _solve_coefficients(
    data: Array, operator: lx.TridiagonalLinearOperator, h=1
) -> Array:
    # Calcualte second and second last coefficients
    c2 = 1 / 6 * data[0]
    cnp2 = 1 / 6 * data[-1]

    # Solve for internal cofficients
    yvec = _construct_vector(data, c2, cnp2)
    solution = lx.linear_solve(operator, yvec)
    cs = solution.value

    # Calculate first and last coefficients
    c1 = 2 * c2 - cs[0]
    cnp3 = 2 * cnp2 - cs[-1]
    return jnp.concatenate([jnp.array([c1, c2]), cs, jnp.array([cnp2, cnp3])])


def _spline_basis(t: Array) -> Array:
    at = jnp.abs(t)
    fn1 = lambda t: (2 - t) ** 3
    fn2 = lambda t: 4 - 6 * t**2 + 3 * t**3
    return jnp.where(
        at >= 1,
        jnp.where(at <= 2, fn1(at), 0),  # type: ignore
        jnp.where(at <= 1, fn2(at), 0),  # type: ignore
    )


def _spline_value(
    coefficients: Array,
    coordinate: Array,
    index: Array,
    mode: str,
    cval: complex | float,
) -> Array:
    coefficient = coefficients.at[tuple(index)].get(mode=mode, fill_value=cval)
    fn = vmap(lambda x, i: _spline_basis(x - i + 1), (0, 0))
    return coefficient * fn(coordinate, index).prod()


def _spline_point(
    coefficients: Array,
    coordinate: Array,
    mode: str,
    cval: complex | float,
) -> Array:
    index_fn = lambda x: (jnp.arange(0, 4) + jnp.floor(x)).astype(int)
    index_vals = vmap(index_fn)(coordinate)
    indices = jnp.array(jnp.meshgrid(*index_vals, indexing="ij"))
    fn = lambda index: _spline_value(coefficients, coordinate, index, mode, cval)
    return vmap(fn)(indices.reshape(coefficients.ndim, -1).T).sum()


#
# Utilities
#
def _unzip2(xys):
    xs = []
    ys = []
    for x, y in xys:
        xs.append(x)
        ys.append(y)
    return tuple(xs), tuple(ys)



"""
Image normalization routines.
"""

import math

import jax.numpy as jnp
from jaxtyping import Array, Bool, Complex, Float, Inexact

from ..jax_util import FloatLike, NDArrayLike


def rescale_image(
    image: Inexact[NDArrayLike, "y_dim x_dim"],
    mean: FloatLike = 0.0,
    std: FloatLike = 1.0,
    *,
    where: Bool[Array, "y_dim x_dim"] | None = None,
) -> Inexact[Array, "y_dim x_dim"]:
    """Normalize so that the image is mean `mean`
    and standard deviation `std` in real space.

    **Parameters:**

    - `image`:
        The image in real-space.
    - `mean`:
        Rescale to this mean.
    - `std`:
        Rescale to this standard deviation.
    - `where`:
        As `where` in `jax.numpy.std` and
        `jax.numpy.mean`. This argument is ignored if
        `input_is_real_space = False`.


    **Returns:**

    Image rescaled to the given mean and standard deviation.
    """
    image = jnp.asarray(image)
    mean, std = jnp.asarray(mean), jnp.asarray(std)
    # First normalize image to zero mean and unit standard deviation
    normalized_image = (image - jnp.mean(image, where=where)) / jnp.std(
        image, where=where
    )
    # Then rescale
    rescaled_image = std * normalized_image + mean

    return rescaled_image


def standardize_image(
    image: Inexact[NDArrayLike, "y_dim x_dim"],
    *,
    where: Bool[Array, "y_dim x_dim"] | None = None,
) -> Inexact[Array, "y_dim x_dim"]:
    """Normalize so that the image is mean 0
    and standard deviation 1 in real space.

    **Parameters:**

    - `image`:
        The image in real-space.
    - `where`:
        As `where` in `jax.numpy.std` and
        `jax.numpy.mean`. This argument is ignored if
        `input_is_real_space = False`.


    **Returns:**

    The standardized image in real-space.
    """
    image = jnp.asarray(image)
    return (image - jnp.mean(image, where=where)) / jnp.std(image, where=where)


def rescale_fft(
    fourier_image: Complex[NDArrayLike, "y_dim x_dim"],
    mean: FloatLike = 0.0,
    std: FloatLike = 1.0,
    *,
    input_is_rfft: bool = True,
    shape_in_real_space: tuple[int, int] | None = None,
) -> Complex[Array, "y_dim x_dim"]:
    """Rescale so that the image is mean `mean`
    and standard deviation `std` in real-space.

    **Parameters:**

    - `fourier_image`:
        The image in the Fourier domain.
        Ensure the zero frequency component is in the corner.
    - `std`:
        The real-space image is rescaled to this standard deviation.
    - `mean`:
        The real-space image is rescaled to this mean.


    **Returns:**

    Image in the Fourier domain rescaled to
    the given mean and standard deviation.
    """
    fourier_image = jnp.asarray(fourier_image)
    n1, n2 = fourier_image.shape
    n_pixels = (
        (
            n1 * (2 * n2 - 1)
            if shape_in_real_space is None
            else math.prod(shape_in_real_space)
        )
        if input_is_rfft
        else n1 * n2
    )
    fourier_image_zero_mean = fourier_image.at[0, 0].set(0.0)
    image_std = (
        jnp.sqrt(
            jnp.sum(jnp.abs(fourier_image_zero_mean[:, 0]) ** 2)
            + 2 * jnp.sum(jnp.abs(fourier_image_zero_mean[:, 1:]) ** 2)
        )
        if input_is_rfft
        else jnp.linalg.norm(fourier_image_zero_mean)
    ) / n_pixels
    normalized_image = fourier_image_zero_mean / image_std
    rescaled_image = (normalized_image * std).at[0, 0].set(mean * n_pixels)

    return rescaled_image


def standardize_fft(
    fourier_image: Complex[NDArrayLike, "y_dim x_dim"],
    *,
    input_is_rfft: bool = True,
    shape_in_real_space: tuple[int, int] | None = None,
) -> Complex[Array, "y_dim x_dim"]:
    """Standardize so that the image is mean 0
    and standard deviation 1 in real-space.

    **Parameters:**

    - `fourier_image`:
        The image in the Fourier domain.
        Ensure the zero frequency component is in the corner.


    **Returns:**

    Standardized image in the Fourier domain.
    """
    return rescale_fft(
        fourier_image,
        mean=0.0,
        std=1.0,
        input_is_rfft=input_is_rfft,
        shape_in_real_space=shape_in_real_space,
    )


def background_subtract_image(image: Float[NDArrayLike, "y_dim x_dim"]):
    """Ensure an image is on a background with mode equal to zero
    by subtracting the mean value on its outer edge.

    Assumes the signal in the image has sufficiently decayed out
    toward the edges.

    **Arguments:**

    - `image`:
        The image to be background subtracted.

    **Returns:**

    The background subtracted image.
    """
    return jnp.asarray(image) - compute_edge_value(image)


def compute_edge_value(image: Float[NDArrayLike, "y_dim x_dim"]):
    """Compute the median of the values at the image edges.
    Useful for background subtraction.

    **Arguments:**

    - `image`:
        The image to retrieve the edge value of.

    **Returns:**

    The median edge value.
    """
    image = jnp.asarray(image)
    edge_values = jnp.concatenate(
        (image[0, :], image[-1, :], image[1:-1, 0], image[1:-1, -1]), axis=0
    )
    return jnp.median(edge_values)



"""
Image formation models simulated from gaussian noise distributions.
"""

from abc import abstractmethod
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..._internal import error_if_not_positive
from ...jax_util import FloatLike
from ...ndimage import (
    AbstractFilter,
    AbstractFourierOperator,
    AbstractMask,
    FourierConstant,
    rfftn,
)
from .._image_model import AbstractImageModel
from .base_noise_model import AbstractEmpiricalNoiseModel, AbstractLikelihoodNoiseModel


RealImageArray = Float[
    Array,
    "{self.image_model.image_config.y_dim} {self.image_model.image_config.x_dim}",  # noqa: E501
]
FourierImageArray = Complex[
    Array,
    "{self.image_model.image_config.y_dim} {self.image_model.image_config.x_dim//2+1}",  # noqa: E501
]
ImageArray = RealImageArray | FourierImageArray


class AbstractGaussianNoiseModel(
    AbstractEmpiricalNoiseModel, AbstractLikelihoodNoiseModel, strict=True
):
    r"""An `AbstractNoiseModel` where images are formed via additive
    gaussian noise.

    Subclasses may compute the likelihood in real or fourier space and
    make different assumptions about the variance / covariance.
    """

    image_model: eqx.AbstractVar[AbstractImageModel]
    signal_scale_factor: eqx.AbstractVar[Float[Array, ""]]
    signal_offset: eqx.AbstractVar[Float[Array, ""]]

    @override
    def sample(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> ImageArray:
        """Sample a noisy image from the gaussian noise model.

        !!! info
            A random number generator key will *not* be passed to
            `AbstractImageModel.simulate`, therefore any
            stochastic elements to the `AbstractImageModel`
            will not be used.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the signal in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        return self.compute_signal(
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        ) + self.compute_noise(
            rng_key,
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

    @override
    def compute_signal(
        self,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> ImageArray:
        """Render the signal from the image formation model.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the signal in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        simulated_image = self.image_model.simulate(
            outputs_real_space=True, mask=None, filter=filter
        )
        simulated_image = (
            error_if_not_positive(self.signal_scale_factor) * simulated_image
            + self.signal_offset
        )
        if mask is not None:
            simulated_image = mask(simulated_image)
        return simulated_image if outputs_real_space else rfftn(simulated_image)

    @abstractmethod
    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> ImageArray:
        """Draw a realization from the gaussian noise model and return either in
        real or fourier space.
        """
        raise NotImplementedError


class GaussianWhiteNoiseModel(AbstractGaussianNoiseModel, strict=True):
    r"""A gaussian noise model, where each pixel is independently drawn from
    a zero-mean gaussian of fixed variance (white noise).

    This computes the likelihood in real space, where the variance is a
    constant value across all pixels.
    """

    image_model: AbstractImageModel
    variance: Float[Array, ""]
    signal_scale_factor: Float[Array, ""]
    signal_offset: Float[Array, ""]

    def __init__(
        self,
        image_model: AbstractImageModel,
        variance: FloatLike = 1.0,
        signal_scale_factor: FloatLike = 1.0,
        signal_offset: FloatLike = 0.0,
    ):
        """**Arguments:**

        - `image_model`:
            The image formation model.
        - `variance`:
            The variance of each pixel.
        - `signal_scale_factor`:
            A scale factor for the underlying signal simulated
            from `image_model`.
        - `signal_offset`:
            An offset for the underlying signal simulated from `image_model`.
        """  # noqa: E501
        self.image_model = image_model
        self.variance = jnp.asarray(variance, dtype=float)
        self.signal_scale_factor = jnp.asarray(signal_scale_factor, dtype=float)
        self.signal_offset = jnp.asarray(signal_offset, dtype=float)

    @override
    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> ImageArray:
        """Sample a realization of the noise from the distribution.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the noise in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        n_pixels = self.image_model.image_config.padded_n_pixels
        frequency_grid = self.image_model.image_config.get_frequency_grid(
            padding=True, physical=True
        )
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(n_pixels * error_if_not_positive(self.variance))
        model = eqx.tree_at(
            lambda x: (x.image_transform, x.normalizes_signal),
            self.image_model,
            (None, False),
            is_leaf=lambda x: x is None,
        )
        noise = model.postprocess(
            std
            * jr.normal(rng_key, shape=frequency_grid.shape[0:-1])
            .at[0, 0]
            .set(0.0)
            .astype(complex),
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

        return noise

    @override
    def log_likelihood(
        self,
        observed: RealImageArray,
        *,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        !!! info

            When computing the likelihood, the observed image is assumed to have already
            been preprocessed with `filter` and `mask`. In other words,
            if `filter` or `mask` are `None`, these
            will *not* be applied to `observed`. The user must do this
            manually if desired.

        **Arguments:**

        - `observed` : The observed data in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        variance = self.variance
        # Create simulated data
        simulated = self.compute_signal(outputs_real_space=True, mask=mask, filter=filter)
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_pixel = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each pixel.
        log_likelihood_per_pixel = -1.0 * (
            squared_standard_normal_per_pixel + jnp.log(2 * jnp.pi * variance) / 2
        )
        # Compute log-likelihood, summing over pixels
        log_likelihood = jnp.sum(log_likelihood_per_pixel)

        return log_likelihood


class GaussianColoredNoiseModel(AbstractGaussianNoiseModel, strict=True):
    r"""A gaussian noise model, where pixels are correlated, but each
    frequency is independent (colored noise).

    This computes the likelihood in Fourier space,
    so that the variance to be an arbitrary noise power spectrum.
    """

    image_model: AbstractImageModel
    variance_fn: AbstractFourierOperator
    signal_scale_factor: Float[Array, ""]
    signal_offset: Float[Array, ""]

    def __init__(
        self,
        image_model: AbstractImageModel,
        variance_fn: AbstractFourierOperator | None = None,
        signal_scale_factor: FloatLike = 1.0,
        signal_offset: FloatLike = 0.0,
    ):
        """**Arguments:**

        - `image_model`:
            The image formation model.
        - `variance_fn`:
            The variance of each fourier mode. By default,
            `cryojax.ndimage.FourierConstant(1.0)`.
        - `signal_scale_factor`:
            A scale factor for the underlying signal simulated from `image_model`.
        - `signal_offset`:
            An offset for the underlying signal simulated from `image_model`.
        """  # noqa: E501
        self.image_model = image_model
        self.variance_fn = variance_fn or FourierConstant(1.0)
        self.signal_scale_factor = jnp.asarray(signal_scale_factor, dtype=float)
        self.signal_offset = jnp.asarray(signal_offset, dtype=float)

    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> ImageArray:
        """Sample a realization of the noise from the distribution.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the noise in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        n_pixels = self.image_model.image_config.padded_n_pixels
        frequency_grid = self.image_model.image_config.get_frequency_grid(
            padding=True, physical=True
        )
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(n_pixels * self.variance_fn(frequency_grid))
        model = eqx.tree_at(
            lambda x: (x.image_transform, x.normalizes_signal),
            self.image_model,
            (None, False),
            is_leaf=lambda x: x is None,
        )
        noise = model.postprocess(
            std
            * jr.normal(rng_key, shape=frequency_grid.shape[0:-1])
            .at[0, 0]
            .set(0.0)
            .astype(complex),
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

        return noise

    @override
    def log_likelihood(
        self,
        observed: FourierImageArray,
        *,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        !!! info

            When computing the likelihood, the observed image is assumed to have already
            been preprocessed with `filter` and `mask`. In other words,
            if `filter` or `mask` are `None`, these
            will *not* be applied to `observed`. The user must do this
            manually if desired.

        **Arguments:**

        - `observed` : The observed data in fourier space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        n_pixels = self.image_model.image_config.n_pixels
        frequency_grid = self.image_model.image_config.get_frequency_grid(
            padding=False, physical=True
        )
        # Compute the variance and scale up to be independent of the number of pixels
        variance = n_pixels * self.variance_fn(frequency_grid)
        # Create simulated data
        simulated = self.compute_signal(
            outputs_real_space=False,
            mask=mask,
            filter=filter,
        )
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_mode = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each fourier mode.
        log_likelihood_per_mode = (
            squared_standard_normal_per_mode + jnp.log(2 * jnp.pi * variance) / 2
        )
        # Compute log-likelihood, throwing away the zero mode. Need to take care
        # to compute the loss function in fourier space for a real-valued function.
        log_likelihood = (
            -1.0
            * (
                jnp.sum(log_likelihood_per_mode[1:, 0])
                + 2 * jnp.sum(log_likelihood_per_mode[:, 1:])
            )
            / n_pixels
        )

        return log_likelihood



from . import (
    atom_util as atom_util,
    constants as constants,
    experimental as experimental,
    io as io,
    jax_util as jax_util,
    ndimage as ndimage,
    rotations as rotations,
    simulator as simulator,
)
from .cryojax_version import __version__ as __version__



"""
Routines for dealing with image cropping and padding.
"""

from typing import Any

import jax
import jax.numpy as jnp
from jaxtyping import Array, Inexact, Int

from ..jax_util import NDArrayLike


def crop_to_shape(
    image_or_volume: (
        Inexact[NDArrayLike, "y_dim x_dim"] | Inexact[NDArrayLike, "z_dim y_dim x_dim"]
    ),
    shape: tuple[int, int] | tuple[int, int, int],
    center: (
        tuple[int, int]
        | tuple[int, int, int]
        | tuple[Int[NDArrayLike, ""], Int[NDArrayLike, ""]]
        | tuple[Int[NDArrayLike, ""], Int[NDArrayLike, ""], Int[NDArrayLike, ""]]
        | None
    ) = None,
) -> (
    Inexact[Array, " {shape[0]} {shape[1]}"]
    | Inexact[Array, " {shape[0]} {shape[1]} {shape[2]}"]
):
    """Crop an image or volume to a new shape around its
    center.

    !!! info

        The behavior of this function is slightly different
        depending on the type of `center`. If `center` is
        a tuple of `jax.Array`s, wraps `jax.lax.dynamic_slice`.
        If `center` is a tuple of `int` or `numpy.ndarrays`,
        crops with array indexing and throws an error if
        the resulting shape isn't equal to the requested shape.
    """
    ndim = image_or_volume.ndim
    if ndim not in [2, 3]:
        raise ValueError(
            "`cryojax.ndimage.crop_to_shape` can only crop images "
            f" and volumes, but got array shape of {image_or_volume.shape}."
        )
    if len(shape) != len(image_or_volume.shape):
        raise ValueError(
            "Mismatch between number of dimensions of desired "
            "crop shape and image or volume shape. "
            f"Desired crop shape was {shape} and "
            f"image or volume shape was {image_or_volume.shape}."
        )
    if center is None:
        center = tuple(n // 2 for n in image_or_volume.shape[::-1])  # type: ignore
        assert center is not None
    if any(isinstance(x, Array) for x in center):
        start_indices = tuple(
            jnp.asarray(x - s // 2) for x, s in zip(center[::-1], shape)
        )
        cropped = jax.lax.dynamic_slice(image_or_volume, start_indices, shape)
    else:
        if len(shape) == 2:
            assert len(center) == 2
            image = image_or_volume
            xc, yc = center
            h, w = shape
            x0, y0 = (max(xc - w // 2, 0), max(yc - h // 2, 0))
            xn, yn = (
                min(xc + w // 2 + w % 2, image.shape[1]),
                min(yc + h // 2 + h % 2, image.shape[0]),
            )
            cropped = image[y0:yn, x0:xn]
        else:
            assert len(center) == 3
            volume = image_or_volume
            xc, yc, zc = center
            d, h, w = shape
            x0, y0, z0 = (max(xc - w // 2, 0), max(yc - h // 2, 0), max(zc - d // 2, 0))
            xn, yn, zn = (
                min(xc + w // 2 + w % 2, volume.shape[2]),
                min(yc + h // 2 + h % 2, volume.shape[1]),
                min(zc + d // 2 + d % 2, volume.shape[0]),
            )
            cropped = volume[z0:zn, y0:yn, x0:xn]
    if cropped.shape != tuple(shape):
        raise ValueError(
            f"The cropped shape {cropped.shape} was not equal to the desired shape "
            f"{shape} in "
            "`cryojax.ndimage.crop_to_shape`. This can happen if the crop is "
            "near the image edges."
        )
    return jnp.asarray(cropped)


def pad_to_shape(
    image_or_volume: (
        Inexact[NDArrayLike, "y_dim x_dim"] | Inexact[NDArrayLike, "z_dim y_dim x_dim"]
    ),
    shape: tuple[int, int] | tuple[int, int, int],
    **kwargs: Any,
) -> (
    Inexact[Array, " {shape[0]} {shape[1]}"]
    | Inexact[Array, " {shape[0]} {shape[1]} {shape[2]}"]
):
    """Pad an image or volume to a new shape."""
    if image_or_volume.ndim not in [2, 3]:
        raise ValueError(
            "pad_to_shape can only pad images and volumes. Got array shape "
            f"of {image_or_volume.shape}."
        )
    if len(shape) != len(image_or_volume.shape):
        raise ValueError(
            "Mismatch between ndim of desired shape and "
            f"array shape. Got a shape of {shape} after padding and "
            f"an array shape of {image_or_volume.shape}."
        )
    if len(shape) == 2:
        image = image_or_volume
        y_padding = _get_left_vs_right_pad(shape[0], image.shape[0])
        x_padding = _get_left_vs_right_pad(shape[1], image.shape[1])
        padding = (y_padding, x_padding)
    elif len(shape) == 3:
        volume = image_or_volume
        z_padding = _get_left_vs_right_pad(shape[0], volume.shape[0])
        y_padding = _get_left_vs_right_pad(shape[1], volume.shape[1])
        x_padding = _get_left_vs_right_pad(shape[2], volume.shape[2])
        padding = (z_padding, y_padding, x_padding)
    else:
        raise ValueError(
            f"pad_to_shape can only pad images and volumes. Got desired shape of {shape}."
        )
    return jnp.pad(image_or_volume, padding, **kwargs)


def resize_with_crop_or_pad(
    image: Inexact[NDArrayLike, "y_dim x_dim"], shape: tuple[int, int], **kwargs
) -> Inexact[Array, " {shape[0]} {shape[1]}"]:
    """Resize an image to a new shape using padding and cropping."""
    if image.ndim != 2 or len(shape) != 2:
        raise ValueError(
            "resize_with_crop_or_pad can only resize images. Got array shape "
            f"of {image.shape} and desired shape {shape}."
        )
    N1, N2 = image.shape
    M1, M2 = shape
    if N1 >= M1 and N2 >= M2:
        image = crop_to_shape(image, shape)
    elif N1 <= M1 and N2 <= M2:
        image = pad_to_shape(image, shape, **kwargs)
    elif N1 <= M1 and N2 >= M2:
        image = crop_to_shape(image, (N1, M2))
        image = pad_to_shape(image, (M1, M2), **kwargs)
    else:
        image = crop_to_shape(image, (M1, N2))
        image = pad_to_shape(image, (M1, M2), **kwargs)

    return image


def _get_left_vs_right_pad(pad_dim, image_dim):
    pad = pad_dim - image_dim
    if pad_dim % 2 == 0 and image_dim % 2 == 0:
        pad_l, pad_r = (pad // 2, pad // 2)
    elif pad_dim % 2 == 1 and image_dim % 2 == 0:
        pad_l, pad_r = (pad // 2, pad // 2 + 1)
    elif pad_dim % 2 == 0 and image_dim % 2 == 1:
        pad_l, pad_r = (pad // 2 + 1, pad // 2)
    else:
        pad_l, pad_r = (pad // 2, pad // 2)
    return pad_l, pad_r



from collections.abc import Callable
from typing import TypeVar

import equinox as eqx
from jaxtyping import ArrayLike, Bool, PyTree

from .._config import CRYOJAX_ENABLE_CHECKS


T = TypeVar("T", bound="PyTree")


def maybe_error_if(
    x: T, pred_fn: Callable[[T], Bool[ArrayLike, "..."]], msg: str
) -> PyTree:
    """Applies [`equinox.error_if`](https://docs.kidger.site/equinox/api/errors/#equinox.error_if)
    depending on the value of the environmental variable `CRYOJAX_ENABLE_CHECKS`.

    - If `CRYOJAX_ENABLE_CHECKS=true`:
        This function is equivalent to `equinox.error_if`, with the replacement of
        the input `pred` with `pred_fn`, where `pred = pred_fn(x)`. This way, `pred` is only evaluated
        if checks are enabled.
    - If `CRYOJAX_ENABLE_CHECKS=false`:
        This function is the identity, i.e. `lambda x: x`.

    By default, `CRYOJAX_ENABLE_CHECKS=false` because checks may cause slowdowns, particularly
    on GPU.

    This function is used to achieve a similar idea as
    ['JAX_ENABLE_CHECKS'](https://docs.jax.dev/en/latest/config_options.html#jax_enable_checks)
    in `cryojax` and is exposed as public API for development downstream.
    """  # noqa: E501
    # `enable_checks` keyword is included for unit testing; it is not a public
    # argument.
    if CRYOJAX_ENABLE_CHECKS:
        return eqx.error_if(x, pred_fn(x), msg)
    else:
        return x



from typing import Any, ClassVar
from typing_extensions import override

import jax.numpy as jnp
from jaxtyping import Array, Float

from ...jax_util import FloatLike
from .._image_config import AbstractImageConfig
from .base_volume import (
    AbstractVolumeIntegrator,
    AbstractVolumeRenderFn,
    AbstractVolumeRepresentation,
    ProjectionArray,
    VoxelArray,
)
from .fourier_voxels import (
    FourierSliceExtraction,
    FourierVoxelGridVolume,
    FourierVoxelSplineVolume,
)
from .gaussian_volume import (
    GaussianMixtureProjection,
    GaussianMixtureRenderFn,
    GaussianMixtureVolume,
)
from .independent_atom_volume import (
    FFTAtomProjection,
    FFTAtomRenderFn,
    IndependentAtomVolume,
)
from .real_voxels import RealVoxelGridVolume, RealVoxelProjection


class AutoVolumeProjection(
    AbstractVolumeIntegrator[AbstractVolumeRepresentation], strict=True
):
    """Volume projection auto selection from cryoJAX
    `AbstractVolumeIntegrator` implementations.

    !!! info
        Based on the [`cryojax.simulator.AbstractVolumeRepresentation`][] passed
        at runtime, this class chooses a default projection method.
        In particular,

        | Volume representation | Projection method | Atom or voxel? |
        | :-------------------- | :------------------ | :------------------ |
        | [`cryojax.simulator.GaussianMixtureVolume`][] | [`cryojax.simulator.GaussianMixtureProjection`][] | atom |
        | [`cryojax.simulator.IndependentAtomVolume`][] | [`cryojax.simulator.FFTAtomProjection`][] | atom |
        | [`cryojax.simulator.FourierVoxelGridVolume`][] or [`cryojax.simulator.FourierVoxelSplineVolume`][] | [`cryojax.simulator.FourierSliceExtraction`][] | voxel |
        | [`cryojax.simulator.RealVoxelGridVolume`][] | [`cryojax.simulator.RealVoxelProjection`][] | voxel |

        To use advanced options for a given projection method,
        see each respective class.

    !!! warning
        If using [`cryojax.simulator.FFTAtomRenderFn`][] or [`cryojax.simulator.RealVoxelProjection`][], [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
        must be installed. See the cryoJAX [installation instructions](https://github.com/michael-0brien/cryojax?tab=readme-ov-file#installation)
        for installing `jax-finufft`.
    """  # noqa: E501

    outputs_ewald_sphere: ClassVar[bool] = False

    def _select_projection_method(
        self, volume: AbstractVolumeRepresentation
    ) -> AbstractVolumeIntegrator:
        if isinstance(volume, (FourierVoxelGridVolume, FourierVoxelSplineVolume)):
            integrator = FourierSliceExtraction()
        elif isinstance(volume, GaussianMixtureVolume):
            integrator = GaussianMixtureProjection()
        elif isinstance(volume, RealVoxelGridVolume):
            integrator = RealVoxelProjection()
        elif isinstance(volume, IndependentAtomVolume):
            integrator = FFTAtomProjection()
        else:
            raise ValueError(
                "Could not use `AutoVolumeProjection` for volume of "
                f"type {type(volume).__name__}. If using a custom volume, "
                "please directly pass an integrator."
            )
        return integrator

    @override
    def integrate(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionArray:
        """Automatically select volume projection method given a
        volume representation.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The image configuration.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in Fourier.

        **Returns:**

        The volume projection in real or Fourier space at the
        `AbstractImageConfig.padded_shape` and the `image_config.pixel_size`.
        """
        volume_integrator = self._select_projection_method(volume_representation)
        return volume_integrator.integrate(
            volume_representation, image_config, outputs_real_space=outputs_real_space
        )


AutoVolumeProjection.__init__.__doc__ = """**Arguments:**

- `options`:
    Keyword arguments passed to the resolved projection method`,
    e.g. `GaussianMixtureProjection(**options)`.
"""


class AutoVolumeRenderFn(
    AbstractVolumeRenderFn[AbstractVolumeRepresentation], strict=True
):
    """Volume rendering auto selection from cryoJAX
    `AbstractVolumeRenderFn` implementations.

    !!! info
        Based on the [`cryojax.simulator.AbstractVolumeRepresentation`][] passed
        at runtime, this class chooses a default rendering function.
        In particular,

        | Volume representation | Rendering function  |
        | :-------------------- | :-----------------  |
        | [`cryojax.simulator.GaussianMixtureVolume`][] | [`cryojax.simulator.GaussianMixtureRenderFn`][] |
        | [`cryojax.simulator.IndependentAtomVolume`][] | [`cryojax.simulator.FFTAtomRenderFn`][] |

        To use advanced options for a given rendering function,
        see each respective class.

    !!! warning
        If using [`cryojax.simulator.FFTAtomRenderFn`][], [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
        must be installed. See the cryoJAX [installation instructions](https://github.com/michael-0brien/cryojax?tab=readme-ov-file#installation)
        for installing `jax-finufft`.
    """  # noqa: E501

    shape: tuple[int, int, int]
    voxel_size: Float[Array, ""]

    options: dict[str, Any]

    def __init__(
        self,
        shape: tuple[int, int, int],
        voxel_size: FloatLike,
        options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            The shape of the voxel grid for rendering.
        - `voxel_size`:
            The voxel size for rendering.
        - `options`:
            Keyword arguments passed to the resolved rendering function,
            e.g. `GaussianMixtureRenderFn(shape, voxel_size, **options)`.
        """
        self.shape = shape
        self.voxel_size = jnp.asarray(voxel_size, dtype=float)
        self.options = options

    def _select_render_method(
        self, volume: AbstractVolumeRepresentation
    ) -> AbstractVolumeRenderFn:
        if isinstance(volume, IndependentAtomVolume):
            return FFTAtomRenderFn(self.shape, self.voxel_size, **self.options)
        elif isinstance(volume, GaussianMixtureVolume):
            return GaussianMixtureRenderFn(self.shape, self.voxel_size, **self.options)
        else:
            raise ValueError(
                "Could not use `AutoVolumeRenderFn` for volume of "
                f"type {type(volume).__name__}. If using a custom volume, "
                "please directly pass its rendering function."
            )

    @override
    def __call__(
        self,
        volume_representation: AbstractVolumeRepresentation,
        *,
        outputs_real_space: bool = True,
        outputs_rfft: bool = False,
        fftshifted: bool = False,
    ) -> VoxelArray:
        render_fn = self._select_render_method(volume_representation)
        return render_fn(
            volume_representation,
            outputs_real_space=outputs_real_space,
            outputs_rfft=outputs_rfft,
            fftshifted=fftshifted,
        )



"""An interface for a grid search method."""

import math
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, Generic

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Int, PyTree

from .custom_types import PyTreeGrid, PyTreeGridPoint, SearchSolution, SearchState
from .pytree_manipulation import (
    tree_grid_shape,
    tree_grid_take,
    tree_grid_unravel_index,
)


class AbstractGridSearchMethod(
    eqx.Module, Generic[SearchState, SearchSolution], strict=True
):
    """An abstract interface that determines the behavior of the grid
    search.
    """

    batch_size: eqx.AbstractVar[int | None]

    @abstractmethod
    def init(
        self,
        tree_grid: PyTreeGrid,
        f_struct: PyTree[jax.ShapeDtypeStruct],
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> SearchState:
        """Initialize the state of the search method.

        **Arguments:**

        - `tree_grid`: As [`run_grid_search`][].
        - `f_struct`: A container that stores the `shape` and `dtype`
                      returned by `fn`.
        - `is_leaf`: As [`run_grid_search`][].

        **Returns:**

        Any pytree that represents the state of the grid search.
        """
        raise NotImplementedError

    @abstractmethod
    def update(
        self,
        fn: Callable[[PyTreeGridPoint, Any], Array],
        tree_grid_point: PyTreeGridPoint,
        args: Any,
        state: SearchState,
        raveled_grid_index: Int[Array, ""],
    ) -> SearchState:
        """Update the state of the grid search.

        **Arguments:**

        - `fn`: As [`run_grid_search`][].
        - `tree_grid_point`: The grid point at which to evaluate `fn`. Specifically,
                             `fn` is evaluated as `fn(tree_grid_point, args)`.
        - `args`: As [`run_grid_search`][].
        - `state`: The current state of the search.
        - `raveled_grid_index`: The current index of the grid. This is
                                used to index `tree_grid` to extract the
                                `tree_grid_point`.

        **Returns:**

        The updated state of the grid search.
        """
        raise NotImplementedError

    @abstractmethod
    def batch_update(
        self,
        fn: Callable[[PyTreeGridPoint, Any], Array],
        tree_grid_point_batch: PyTreeGridPoint,
        args: Any,
        state: SearchState,
        raveled_grid_index_batch: Int[Array, " _"],
    ) -> SearchState:
        """Update the state of the grid search with a batch of grid points as
        input.

        **Arguments:**

        - `fn`: As [`run_grid_search`][].
        - `tree_grid_point_batch`: The grid points at which to evaluate `fn` in
                                   parallel.
        - `args`: As [`run_grid_search`][].
        - `state`: The current state of the search.
        - `raveled_grid_index_batch`: The current batch of indices on which to evaluate
                                      the grid.

        **Returns:**

        The updated state of the grid search.
        """
        raise NotImplementedError

    @abstractmethod
    def postprocess(
        self,
        tree_grid: PyTreeGrid,
        final_state: SearchState,
        f_struct: PyTree[jax.ShapeDtypeStruct],
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> SearchSolution:
        """Post-process the final state of the grid search into a
        solution.

        **Arguments:**

        - `tree_grid`: As [`run_grid_search`][].
        - `final_state`: The final state of the grid search.
        - `f_struct`: A container that stores the `shape` and `dtype`
                      returned by `fn`.
        - `is_leaf`: As [`run_grid_search`][].

        **Returns:**

        Any pytree that represents the solution of the grid search.
        """
        raise NotImplementedError


class _MinimumState(eqx.Module, strict=True):
    current_minimum_eval: Array
    current_best_raveled_index: Array
    current_eval: Array | None = None


class _MinimumSolution(eqx.Module, strict=True):
    value: PyTreeGridPoint | None
    stats: dict[str, Any]
    state: _MinimumState


class MinimumSearchMethod(
    AbstractGridSearchMethod[_MinimumState, _MinimumSolution], strict=True
):
    """Simply find the minimum value returned by `fn` over all grid points.

    The minimization is done *elementwise* for the output returned by `fn(y, args)`.
    This allows for more clever grid searches than a brute-force approach--for example,
    `fn` can explore its own region of parameter space in parallel.
    """

    stores_solution_value: bool
    stores_current_eval: bool
    batch_size: int | None

    def __init__(
        self,
        *,
        stores_solution_value: bool = True,
        stores_current_eval: bool = False,
        batch_size: int | None = None,
    ):
        """**Arguments:**

        - `stores_solution_value`: If `True`, the grid search solution will contain the
                                best grid point found. If `False`, only the flattened
                                index corresponding to these grid points are returned
                                and [`tree_grid_take`][] must be used to extract the
                                actual grid points. Setting this to `False` may be
                                necessary if the grid contains large arrays.
        - `stores_current_eval`: If `True`, carry over the last function evaluation in
                                the `_MinimumState`. This is useful when wrapping this
                                class into new `AbstractGridSearchMethod`s.
        - `batch_size`: The stride of grid points over which to evaluate in parallel.
        """
        self.stores_solution_value = stores_solution_value
        self.stores_current_eval = stores_current_eval
        self.batch_size = batch_size

    def init(
        self,
        tree_grid: PyTreeGrid,
        f_struct: PyTree[jax.ShapeDtypeStruct],
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> _MinimumState:
        # Initialize the state, just keeping track of the best function values
        # and their respective grid index
        return _MinimumState(
            current_minimum_eval=jnp.full(f_struct.shape, jnp.inf, dtype=float),
            current_best_raveled_index=jnp.full(f_struct.shape, 0, dtype=int),
            current_eval=(
                (
                    jnp.full(f_struct.shape, 0.0, dtype=float)
                    if self.batch_size is None
                    else jnp.full((self.batch_size, *f_struct.shape), 0.0, dtype=float)
                )
                if self.stores_current_eval
                else None
            ),
        )

    def update(
        self,
        fn: Callable[[PyTreeGridPoint, Any], Array],
        tree_grid_point: PyTreeGridPoint,
        args: Any,
        state: _MinimumState,
        raveled_grid_index: Int[Array, ""],
    ) -> _MinimumState:
        # Evaluate the function
        value = fn(tree_grid_point, args)
        # Unpack the current state
        last_minimum_value = state.current_minimum_eval
        last_best_raveled_index = state.current_best_raveled_index
        # Update the minimum and best grid index, elementwise
        is_less_than_last_minimum = value < last_minimum_value
        current_minimum_eval = jnp.where(
            is_less_than_last_minimum, value, last_minimum_value
        )
        current_best_raveled_index = jnp.where(
            is_less_than_last_minimum, raveled_grid_index, last_best_raveled_index
        )
        return _MinimumState(
            current_minimum_eval,
            current_best_raveled_index,
            current_eval=value if self.stores_current_eval else None,
        )

    def batch_update(
        self,
        fn: Callable[[PyTreeGridPoint, Any], Array],
        tree_grid_point_batch: PyTreeGridPoint,
        args: Any,
        state: _MinimumState,
        raveled_grid_index_batch: Int[Array, " _"],
    ) -> _MinimumState:
        # Evaluate the batch of grid points and extract the best one
        value_batch = jax.vmap(fn, in_axes=[0, None])(tree_grid_point_batch, args)
        best_batch_index = jnp.argmin(value_batch, axis=0)
        raveled_grid_index = jnp.take(raveled_grid_index_batch, best_batch_index)
        value = jnp.amin(value_batch, axis=0)
        # Unpack the current state
        last_minimum_value = state.current_minimum_eval
        last_best_raveled_index = state.current_best_raveled_index
        # Update the minimum and best grid index, elementwise
        is_less_than_last_minimum = value < last_minimum_value
        current_minimum_eval = jnp.where(
            is_less_than_last_minimum, value, last_minimum_value
        )
        current_best_raveled_index = jnp.where(
            is_less_than_last_minimum, raveled_grid_index, last_best_raveled_index
        )
        return _MinimumState(
            current_minimum_eval,
            current_best_raveled_index,
            current_eval=value_batch if self.stores_current_eval else None,
        )

    def postprocess(
        self,
        tree_grid: PyTreeGrid,
        final_state: _MinimumState,
        f_struct: PyTree[jax.ShapeDtypeStruct],
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> _MinimumSolution:
        # Make sure that shapes did not get modified during loop
        if final_state.current_best_raveled_index.shape != f_struct.shape:
            raise ValueError(
                "The shape of the search state solution does "
                "not match the shape of the output of `fn`. Got "
                f"output shape {f_struct.shape} for `fn`, but got "
                f"shape {final_state.current_best_raveled_index.shape} for the "
                "solution."
            )
        if self.stores_solution_value:
            # Extract the solution of the search, i.e. the grid point(s) corresponding
            # to the raveled grid index
            if f_struct.shape == ():
                raveled_index = final_state.current_best_raveled_index
            else:
                raveled_index = final_state.current_best_raveled_index.ravel()
            # ... get the pytree representation of the index
            tree_grid_index = tree_grid_unravel_index(
                raveled_index, tree_grid, is_leaf=is_leaf
            )
            # ... index the full grid, reshaping the solution's leaves to be the same
            # shape as returned by `fn`
            _reshape_fn = lambda x: (
                x.reshape((*f_struct.shape, *x.shape[1:]))
                if x.ndim > 1
                else x.reshape(f_struct.shape)
            )
            value = jtu.tree_map(_reshape_fn, tree_grid_take(tree_grid, tree_grid_index))
        else:
            value = None
        # ... build and return the solution
        return _MinimumSolution(
            value,
            {"grid_size": math.prod(tree_grid_shape(tree_grid, is_leaf=is_leaf))},
            final_state,
        )



"""This is a module of utilities for working with JAX/Equinox. This includes utilities
for Equinox filtered transformations and Equinox recommendations for creating custom
per-leaf behavior for pytrees.
"""

from ._batched_loop import filter_bmap as filter_bmap, filter_bscan as filter_bscan
from ._errors import maybe_error_if as maybe_error_if
from ._filter_specs import make_filter_spec as make_filter_spec
from ._grid_search import (
    AbstractGridSearchMethod as AbstractGridSearchMethod,
    MinimumSearchMethod as MinimumSearchMethod,
    run_grid_search as run_grid_search,
    tree_grid_shape as tree_grid_shape,
    tree_grid_take as tree_grid_take,
    tree_grid_unravel_index as tree_grid_unravel_index,
)
from ._linear_operator import make_linear_operator as make_linear_operator
from ._pytree_transforms import (
    AbstractPyTreeTransform as AbstractPyTreeTransform,
    CustomTransform as CustomTransform,
    NonArrayStaticTransform as NonArrayStaticTransform,
    StopGradientTransform as StopGradientTransform,
    resolve_transforms as resolve_transforms,
)
from ._typing import (
    BoolLike as BoolLike,
    ComplexLike as ComplexLike,
    FloatLike as FloatLike,
    InexactLike as InexactLike,
    IntLike as IntLike,
    NDArrayLike as NDArrayLike,
)



import os

import jax
import pytest


# jax.config.update("jax_numpy_dtype_promotion", "strict")
# jax.config.update("jax_numpy_rank_promotion", "raise")
jax.config.update("jax_enable_x64", True)


@pytest.fixture
def sample_mrc_path():
    return os.path.join(os.path.dirname(__file__), "data", "3j9g_potential_ps4_4.mrc")


@pytest.fixture
def sample_pdb_path():
    return os.path.join(os.path.dirname(__file__), "data", "1uao.pdb")


@pytest.fixture
def sample_pdb_path_assembly():
    return os.path.join(os.path.dirname(__file__), "data", "1uao_assembly.pdb")


@pytest.fixture
def sample_cif_path():
    return os.path.join(os.path.dirname(__file__), "data", "8zpm.cif")


@pytest.fixture
def sample_waterbox_pdb():
    return os.path.join(os.path.dirname(__file__), "data", "relaxed_small_box_tip3p.pdb")


@pytest.fixture
def peng_parameters_path():
    return os.path.join(os.path.dirname(__file__), "data", "peng1996_params.npy")


@pytest.fixture
def lobato_parameters_path():
    return os.path.join(os.path.dirname(__file__), "data", "lobato2014_params.npy")



import cryojax.ndimage as im
import cryojax.simulator as cxs
import equinox as eqx
import jax.numpy as jnp
import numpy as np
import pytest
from cryojax.io import read_array_from_mrc, read_atoms_from_pdb
from cryojax.ndimage import (
    CircularCosineMask,
    LowpassFilter,
    ScaleImage,
    crop_to_shape,
    pad_to_shape,
)


@pytest.fixture
def pdb_info(sample_pdb_path):
    return read_atoms_from_pdb(sample_pdb_path, center=True, loads_properties=True)


@pytest.fixture
def voxel_info(sample_mrc_path):
    return read_array_from_mrc(sample_mrc_path, loads_grid_spacing=True)


@pytest.fixture
def voxel_volume(voxel_info):
    return cxs.FourierVoxelGridVolume.from_real_voxel_grid(voxel_info[0], pad_scale=1.3)


@pytest.fixture
def voxel_size(voxel_info):
    return voxel_info[1]


@pytest.fixture
def basic_config(voxel_volume, voxel_size):
    shape = voxel_volume.shape[0:2]
    return cxs.BasicImageConfig(
        shape=(int(0.9 * shape[0]), int(0.9 * shape[1])),
        pixel_size=voxel_size,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=shape),
    )


@pytest.fixture
def image_model(voxel_volume, basic_config):
    return cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=cxs.EulerAnglePose(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AstigmaticCTF()),
    )


# Test correct image shape
@pytest.mark.parametrize("model", ["image_model"])
def test_real_shape(model, request):
    """Make sure shapes are as expected in real space."""
    model = request.getfixturevalue(model)
    image = model.simulate(outputs_real_space=True)
    padded_image = model.raw_simulate(outputs_real_space=True)
    assert image.shape == model.image_config.shape
    assert padded_image.shape == model.image_config.padded_shape


@pytest.mark.parametrize("model", ["image_model"])
def test_fourier_shape(model, request):
    """Make sure shapes are as expected in fourier space."""
    model = request.getfixturevalue(model)
    image = model.simulate(outputs_real_space=False)
    padded_image = model.raw_simulate(outputs_real_space=False)
    assert image.shape == model.image_config.get_frequency_grid(padding=False).shape[0:2]
    assert (
        padded_image.shape
        == model.image_config.get_frequency_grid(padding=True).shape[0:2]
    )


@pytest.mark.parametrize("extra_dim_y, extra_dim_x", [(1, 1), (1, 0), (0, 1)])
def test_even_vs_odd_image_shape(extra_dim_y, extra_dim_x, voxel_volume, voxel_size):
    control_shape = voxel_volume.shape[0:2]
    test_shape = (control_shape[0] + extra_dim_y, control_shape[1] + extra_dim_x)
    config_control = cxs.BasicImageConfig(
        control_shape, pixel_size=voxel_size, voltage_in_kilovolts=300.0
    )
    config_test = cxs.BasicImageConfig(
        test_shape, pixel_size=voxel_size, voltage_in_kilovolts=300.0
    )
    pose = cxs.EulerAnglePose()
    transfer_theory = cxs.ContrastTransferTheory(cxs.AstigmaticCTF())
    model_control = cxs.make_image_model(
        voxel_volume, config_control, pose=pose, transfer_theory=transfer_theory
    )
    model_test = cxs.make_image_model(
        voxel_volume, config_test, pose=pose, transfer_theory=transfer_theory
    )
    np.testing.assert_allclose(
        crop_to_shape(model_test.simulate(), control_shape),
        model_control.simulate(),
        atol=1e-3,
    )


@pytest.mark.parametrize(
    "offset_xy, pixel_size, shape, pad_scale",
    (
        ((1, 1), 1.0, (31, 31), 2),
        ((-1, -1), 1.0, (32, 32), 2),
        ((1, -1), 1.0, (31, 32), 2),
        ((-1, 1), 1.0, (32, 31), 2),
    ),
)
def test_translate_mode(pdb_info, offset_xy, pixel_size, shape, pad_scale):
    atom_pos, _, _ = pdb_info
    image_config = cxs.BasicImageConfig(
        shape,
        pixel_size,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=(pad_scale * shape[0], pad_scale * shape[1])),
    )
    gaussian_width = 2 * pixel_size
    volume, integrator = (
        cxs.GaussianMixtureVolume(atom_pos, amplitudes=1.0, variances=gaussian_width**2),
        cxs.GaussianMixtureProjection(),
    )
    pose = cxs.EulerAnglePose.from_translation(np.asarray(offset_xy))
    # Projections
    fft_proj_model = cxs.ProjectionImageModel(
        volume,
        pose,
        image_config,
        integrator,
        translate_mode="fft",
    )
    atom_proj_model = cxs.ProjectionImageModel(
        volume,
        pose,
        image_config,
        integrator,
        translate_mode="atom",
    )
    atom_translate_proj = compute_image(atom_proj_model)
    fft_translate_proj = compute_image(fft_proj_model)

    np.testing.assert_allclose(atom_translate_proj, fft_translate_proj, atol=1e-8)
    # Images
    transfer_theory = cxs.ContrastTransferTheory(cxs.AstigmaticCTF())
    fft_im_model = cxs.LinearImageModel(
        volume,
        pose,
        image_config,
        transfer_theory,
        integrator,
        translate_mode="fft",
    )
    atom_im_model = cxs.LinearImageModel(
        volume,
        pose,
        image_config,
        transfer_theory,
        integrator,
        translate_mode="atom",
    )
    atom_translate_im = compute_image(atom_im_model)
    fft_translate_im = compute_image(fft_im_model)

    np.testing.assert_allclose(atom_translate_im, fft_translate_im, atol=1e-8)


def test_bad_translate_mode(voxel_info, basic_config):
    real_voxels, _ = voxel_info
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxels)
    with pytest.raises(ValueError):
        model = cxs.make_image_model(
            voxel_volume, basic_config, pose=cxs.EulerAnglePose(), translate_mode="atom"
        )
        _ = model.simulate()


@pytest.mark.parametrize(
    "std, signal_centering",
    (
        (2.0, "mean"),
        (4.0, "mean"),
        (2.0, "bg"),
        (4.0, "bg"),
    ),
)
def test_normalize_and_transform(std, signal_centering, voxel_info, basic_config):
    real_voxels, _ = voxel_info
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxels)
    image_model = cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=cxs.EulerAnglePose(),
        image_transform=ScaleImage(scale=std),
        normalizes_signal=True,
        signal_centering=signal_centering,
    )
    image = compute_image(image_model)
    np.testing.assert_approx_equal(np.std(image), std)


@pytest.mark.parametrize("use_transform", (True, False))
def test_mask_zeros_edges(use_transform, voxel_info, basic_config):
    real_voxels, _ = voxel_info
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxels)
    image_model = cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=cxs.EulerAnglePose(),
        image_transform=(ScaleImage(scale=1.0) if use_transform else None),
        normalizes_signal=True,
    )
    image = image_model.simulate(
        mask=CircularCosineMask(
            basic_config.get_coordinate_grid(physical=False), radius=10, rolloff_width=0
        )
    )
    ny, nx = image.shape
    np.testing.assert_allclose(image[0, 0], 0.0)
    np.testing.assert_allclose(image[ny, 0], 0.0)
    np.testing.assert_allclose(image[0, nx], 0.0)
    np.testing.assert_allclose(image[ny, nx], 0.0)


def test_filter_padded_shape(voxel_info, basic_config):
    real_voxels, _ = voxel_info
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxels)
    image_model = cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=cxs.EulerAnglePose(),
    )
    _ = image_model.simulate(
        filter=LowpassFilter(
            basic_config.get_frequency_grid(padding=True, physical=False),
            grid_spacing=basic_config.pixel_size,
        )
    )
    with pytest.raises(ValueError):
        _ = image_model.simulate(
            filter=LowpassFilter(
                basic_config.get_frequency_grid(physical=False),
                grid_spacing=basic_config.pixel_size,
            )
        )


def test_bg_subtract(voxel_info):
    real_voxels, voxel_size = voxel_info
    dim = real_voxels.shape[0]
    padded_dim = 2 * dim
    real_voxels = pad_to_shape(real_voxels, tuple(3 * [padded_dim]))
    image_config = cxs.BasicImageConfig(
        shape=(padded_dim, padded_dim),
        pixel_size=voxel_size,
        voltage_in_kilovolts=300.0,
    )
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxels)
    transfer_theory = cxs.ContrastTransferTheory(cxs.AstigmaticCTF())
    image_model = cxs.make_image_model(
        voxel_volume,
        image_config,
        pose=cxs.EulerAnglePose(),
        transfer_theory=transfer_theory,
        normalizes_signal=True,
        signal_centering="bg",
        quantity_mode="intensity",
    )
    image = compute_image(image_model)
    np.testing.assert_allclose(image[:, 0], 0.0, atol=5e-2)


@pytest.mark.parametrize("rotation_convention", ("object", "frame", "asfsdkl"))
def test_rotation_convention(rotation_convention):
    pose = cxs.EulerAnglePose(phi_angle=45.0, theta_angle=80.0, psi_angle=-30.0)
    volumes = [
        cxs.RealVoxelGridVolume.from_real_voxel_grid(jnp.zeros((10, 10, 10))),
        cxs.FourierVoxelGridVolume.from_real_voxel_grid(jnp.zeros((10, 10, 10))),
        cxs.FourierVoxelSplineVolume.from_real_voxel_grid(jnp.zeros((10, 10, 10))),
        cxs.GaussianMixtureVolume(jnp.zeros((10, 3)), 1.0, 1.0),
        cxs.IndependentAtomVolume(jnp.zeros((10, 3)), im.FourierGaussian()),
    ]
    config = cxs.BasicImageConfig((10, 10), 1.0, 300.0)
    make_wrapper = lambda _v: cxs.make_image_model(
        _v, config, pose, rotation_convention=rotation_convention
    )
    for volume in volumes:
        if rotation_convention in ["object", "frame"]:
            image_model = make_wrapper(volume)
            if volume.rotation_convention == rotation_convention:
                quat1, quat2 = pose.rotation.wxyz, image_model.pose.rotation.wxyz
            else:
                quat1, quat2 = (
                    pose.to_inverse_rotation().rotation.wxyz,
                    image_model.pose.rotation.wxyz,
                )
            np.testing.assert_allclose(quat1, quat2)
        else:
            with pytest.raises(ValueError):
                _ = make_wrapper(volume)


@eqx.filter_jit
def compute_image(image_model):
    return image_model.simulate()



from time import time

import cryojax.ndimage as im
import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from cryojax.constants import PengScatteringFactorParameters
from cryojax.io import read_atoms_from_pdb
from cryojax.rotations import SO3
from jaxtyping import PRNGKeyArray


def setup(num_images, path_to_pdb):
    @eqx.filter_vmap(in_axes=(0, None))
    def make_particle_parameters(key: PRNGKeyArray, config: cxs.BasicImageConfig):
        """Generate random parameters."""
        # Pose
        # ... instantiate rotations
        key, subkey = jax.random.split(
            key
        )  # split the key to use for the next random number
        rotation = SO3.sample_uniform(subkey)

        # ... now in-plane translation
        ny, nx = config.shape

        key, subkey = jax.random.split(key)  # do this everytime you use a key!!
        offset_in_angstroms = (
            jax.random.uniform(subkey, (2,), minval=-0.1, maxval=0.1)
            * jnp.asarray((nx, ny))
            / 2
            * config.pixel_size
        )
        # ... build the pose
        pose = cxs.EulerAnglePose.from_rotation_and_translation(
            rotation, offset_in_angstroms
        )

        # CTF Parameters
        # ... defocus
        key, subkey = jax.random.split(key)
        defocus_in_angstroms = jax.random.uniform(subkey, (), minval=10000, maxval=15000)
        # ... astigmatism
        key, subkey = jax.random.split(key)
        astigmatism_in_angstroms = jax.random.uniform(subkey, (), minval=0, maxval=100)
        key, subkey = jax.random.split(key)
        astigmatism_angle = jax.random.uniform(subkey, (), minval=0, maxval=jnp.pi)
        # Now non-random values
        spherical_aberration_in_mm = 2.7
        amplitude_contrast_ratio = 0.1
        # Build the CTF
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(
                defocus_in_angstroms=defocus_in_angstroms,
                astigmatism_in_angstroms=astigmatism_in_angstroms,
                astigmatism_angle=astigmatism_angle,
                spherical_aberration_in_mm=spherical_aberration_in_mm,
            ),
            amplitude_contrast_ratio=amplitude_contrast_ratio,
        )

        return {
            "image_config": config,
            "pose": pose,
            "transfer_theory": transfer_theory,
        }

    # Generate particle parameters. First, the image config
    pad_options = dict(shape=(200, 200))

    config = cxs.BasicImageConfig(
        shape=(150, 150),
        pixel_size=2.0,
        voltage_in_kilovolts=300.0,
        pad_options=pad_options,
    )
    # ... RNG keys
    keys = jax.random.split(jax.random.key(0), num_images)
    # ... make parameters
    particle_parameters = make_particle_parameters(keys, config)

    atom_positions, atom_types, atom_properties = read_atoms_from_pdb(
        path_to_pdb,
        center=True,
        loads_properties=True,
        selection_string="name CA",  # C-Alphas for simplicity
    )
    scattering_parameters = PengScatteringFactorParameters(atom_types)
    volume_gmm = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        scattering_parameters,
        extra_b_factors=atom_properties["b_factors"],
    )
    render_fn = cxs.GaussianMixtureRenderFn(shape=(200, 200, 200), voxel_size=1.0)
    volume_fourier_grid = cxs.FourierVoxelGridVolume.from_real_voxel_grid(
        render_fn(volume_gmm), pad_scale=2
    )
    atom_volume = cxs.IndependentAtomVolume(
        positions=atom_positions,
        scattering_factors=im.FourierGaussian(amplitude=1.0, b_factor=10.0),
    )
    return (particle_parameters, volume_gmm, atom_volume, volume_fourier_grid)


@eqx.filter_vmap(in_axes=(eqx.if_array(0), eqx.if_array(0), eqx.if_array(0), None, None))
def simulate_image_nojit(
    image_config, pose, transfer_theory, potential, volume_integrator
):
    image_model = cxs.make_image_model(
        volume=potential,
        image_config=image_config,
        pose=pose,
        transfer_theory=transfer_theory,
        volume_integrator=volume_integrator,
    )
    return image_model.simulate()


simulate_image_jit = eqx.filter_jit(simulate_image_nojit)


def run(n_iterations, num_images, path_to_pdb):
    particle_parameters, volume_gmm, atom_volume, volume_fourier_grid = setup(
        num_images, path_to_pdb
    )

    image_config, pose, transfer_theory = (
        particle_parameters["image_config"],
        particle_parameters["pose"],
        particle_parameters["transfer_theory"],
    )
    dim, num_atoms = image_config.padded_y_dim, atom_volume.positions.shape[0]
    print(
        f"Benchmarking simulation of {n_images} {dim}x{dim} images of "
        f"a structure with {num_atoms} atoms..."
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        with jax.disable_jit():
            fft_image = simulate_image_nojit(
                image_config,
                pose,
                transfer_theory,
                atom_volume,
                cxs.FFTAtomProjection(eps=1e-16),
            )
            fft_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fft_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fft_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"NUFFT (no JIT): {1000 * fft_avg_time:.2f} +/- {1000 * fft_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fft_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            atom_volume,
            cxs.FFTAtomProjection(eps=1e-16),
        )
        fft_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fft_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fft_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"NUFFT (JIT): {1000 * fft_avg_time:.2f} +/- {1000 * fft_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        with jax.disable_jit():
            gmm_image = simulate_image_nojit(
                image_config,
                pose,
                transfer_theory,
                volume_gmm,
                cxs.GaussianMixtureProjection(),
            )
            gmm_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
    gmm_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"GMM (no JIT): {1000 * gmm_avg_time:.2f} +/- {1000 * gmm_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        gmm_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            volume_gmm,
            cxs.GaussianMixtureProjection(),
        )
        gmm_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    jit_gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
    jit_gmm_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"GMM (JIT): {1000 * jit_gmm_avg_time:.2f} +/- {1000 * jit_gmm_std_time:.2f} ms"
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        with jax.disable_jit():
            fs_image = simulate_image_nojit(
                image_config,
                pose,
                transfer_theory,
                volume_fourier_grid,
                cxs.FourierSliceExtraction(),
            )
            fs_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fs_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"Fourier Slice (no JIT): {1000 * fs_avg_time:.2f} "
        f"+/- {1000 * fs_std_time:.2f} ms"
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fs_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            volume_fourier_grid,
            cxs.FourierSliceExtraction(),
        )
        fs_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fs_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"Fourier Slice (JIT): {1000 * fs_avg_time:.2f} +/- {1000 * fs_std_time:.2f} ms"
    )


if __name__ == "__main__":
    n_iterations, n_images = 25, 10
    # Necessary to pull from git LFS
    path_to_pdb = "../docs/examples/data/thyroglobulin_initial.pdb"
    run(n_iterations, n_images, path_to_pdb)



"""
Representations of rigid body rotations and translations of 3D coordinate systems.
"""

from abc import abstractmethod
from collections.abc import Sequence
from functools import cached_property
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import AbstractVar, Module
from jaxtyping import Array, Complex, Float

from ..jax_util import FloatLike, NDArrayLike
from ..ndimage import FourierPhaseShifts, enforce_rfftn_self_conjugates
from ..rotations import SO3, convert_quaternion_to_euler_angles


class AbstractPose(Module, strict=True):
    """Base class for the image pose. Subclasses will choose a
    particular convention for parameterizing the rotation by
    overwriting the `AbstractPose.rotation` property.
    """

    offset_in_angstroms: AbstractVar[Float[Array, "2"] | Float[Array, "3"]]

    def rotate_coordinates(
        self,
        coordinate_grid_or_list: (
            Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]
        ),
        inverse: bool = False,
    ) -> Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]:
        """Rotate a 3D coordinate system.

        **Arguments:**

        - `coordinate_grid_or_list`:
            The 3D coordinate system to rotate. This can either be a list of coordinates
            of shape `(N, 3)` or a grid of coordinates `(N1, N2, N3, 3)`.
        - `inverse`:
            If `True`, compute the inverse rotation (i.e. rotation by the matrix $R^T$,
            where $R$ is the rotation matrix).

        **Returns:**

        The rotated version of `coordinate_grid_or_list`.
        """
        rotation = self.rotation.inverse() if inverse else self.rotation
        if isinstance(coordinate_grid_or_list, Float[Array, "size 3"]):  # type: ignore
            rotated_coordinate_grid_or_list = jax.vmap(rotation.apply)(
                coordinate_grid_or_list
            )
        elif isinstance(coordinate_grid_or_list, Float[Array, "z_dim y_dim x_dim 3"]):  # type: ignore
            rotated_coordinate_grid_or_list = jax.vmap(
                jax.vmap(jax.vmap(rotation.apply))
            )(coordinate_grid_or_list)
        else:
            raise ValueError(
                "Coordinates must be a JAX array either of shape (N, 3) or "
                f"(N1, N2, N3, 3). Instead, got {coordinate_grid_or_list.shape} and type "
                f"{type(coordinate_grid_or_list)}."
            )
        return rotated_coordinate_grid_or_list

    def translate_image(
        self,
        fourier_image: Complex[Array, "{shape[0]} {shape[1]}//2+1"],
        translation_operator: Complex[Array, "{shape[0]} {shape[1]}//2+1"],
        shape: tuple[int, int],
    ) -> Complex[Array, "{shape[0]} {shape[1]}//2+1"]:
        """Apply translational phase shifts to a fourier-space image.

        **Arguments:**

        - `fourier_image`:
            The image in fourier-space, which is the output of a call
            to `cryojax.image.rfftn`.
        - `phase_shifts`:
            The phase shifts for translation, which are computed from
            `AbstractPose.compute_translation_operator`.
        - `shape`:
            The shape of `fourier_image` in real-space.

        **Return:**

        The translated `fourier_image`, taking care to avoid image
        artifacts when applying the phase shifts.
        """
        fourier_image = enforce_rfftn_self_conjugates(
            fourier_image, shape, includes_dc=False, mode="zero"
        )
        return fourier_image * translation_operator

    def compute_translation_operator(
        self, frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"]
    ) -> Complex[Array, "y_dim x_dim"]:
        """Compute the phase shifts from the in-plane translation,
        given a frequency grid coordinate system.

        **Arguments:**

        - `frequency_grid_in_angstroms`:
            A grid of in-plane frequency coordinates $(q_x, q_y)$

        **Returns:**

        From the vector $(t_x, t_y)$ (given by `self.offset_in_angstroms`), returns the
        grid of in-plane phase shifts $\\exp{(- 2 \\pi i (t_x q_x + t_y q_y))}$.
        """
        xy = self.offset_in_angstroms[0:2]
        compute_fn = FourierPhaseShifts(xy)
        return compute_fn(frequency_grid_in_angstroms)

    @cached_property
    def offset_x_in_angstroms(self) -> Float[Array, "..."]:
        """The in-plane translation in the x direction."""
        # a[..., i] indexing is for convenience outside of `jax.vmap`
        # regions. Be careful! Will silently cause failures
        # if batch dimensions are not the leading dimensions.
        return self.offset_in_angstroms[..., 0]

    @cached_property
    def offset_y_in_angstroms(self) -> Float[Array, "..."]:
        """The in-plane translation in the y direction."""
        # a[..., i] indexing is for convenience outside of `jax.vmap`
        # regions. Be careful! Will silently cause failures
        # if batch dimensions are not the leading dimensions.
        return self.offset_in_angstroms[..., 1]

    @cached_property
    def offset_z_in_angstroms(self) -> Float[Array, "..."] | None:
        """The out-of-plane translation in the z direction."""
        # a[..., i] is for convenience outside of `jax.vmap`
        # regions. Be careful! Will silently cause failures
        # if batch dimensions are not the leading dimensions.
        return (
            None
            if self.offset_in_angstroms.shape[-1] == 2
            else self.offset_in_angstroms[..., 2]
        )

    @cached_property
    @abstractmethod
    def rotation(self) -> SO3:
        """Generate an `SO3` object from a particular angular
        parameterization.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        """Construct an `AbstractPose` from an `SO3` object."""
        raise NotImplementedError

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        offset_in_angstroms: Float[Array, "2"] | Float[Array, "3"],
    ) -> Self:
        """Construct an `AbstractPose` from an `SO3` object and a
        translation vector.
        """
        if offset_in_angstroms.shape not in [(2,), (3,)]:
            raise ValueError(
                "Array `offset_in_angstroms` given to constructor "
                f"`{cls.__name__}.from_rotation_and_translation` supports "
                "shapes `(2,)` and `(3,)`. Got shape "
                f"`{offset_in_angstroms.shape}`"
            )
        return eqx.tree_at(
            lambda p: p.offset_in_angstroms,
            cls.from_rotation(rotation),
            offset_in_angstroms,
        )

    @classmethod
    def from_translation(
        cls,
        offset_in_angstroms: Float[NDArrayLike, "2"] | Float[NDArrayLike, "3"],
    ) -> Self:
        """Construct an `AbstractPose` from a
        translation vector.
        """
        return eqx.tree_at(
            lambda p: p.offset_in_angstroms,
            cls.from_rotation(SO3(wxyz=jnp.asarray((1, 0, 0, 0), dtype=float))),
            jnp.asarray(offset_in_angstroms, dtype=float),
        )

    @abstractmethod
    def to_inverse_rotation(self) -> Self:
        """Convert an `AbstractPose` to the inverse of its rotation
        representation.
        """
        raise NotImplementedError


class EulerAnglePose(AbstractPose, strict=True):
    r"""An `AbstractPose` represented by Euler angles.
    Angles are given in degrees, and the sequence of rotations is a
    zyz *extrinsic* rotation, with `phi_angle` as the first euler angle,
    `theta_angle` as the second, and `psi_angle` is the third.

    !!! info "Converting to RELION and FREALIGN convention"

        RELION/FREALIGN convention is that the euler angles represent
        a zyz *intrinsic* rotation that "undoes" the rotation in the image. cryoJAX
        defines its convention to be a zyz *extrinsic* rotation that generates the
        pose in the image. In order to convert to the RELION/FREALIGN convention,
        simply **negate each euler angle**.
    """

    offset_in_angstroms: Float[Array, "2"] | Float[Array, "3"]

    phi_angle: Float[Array, ""]
    theta_angle: Float[Array, ""]
    psi_angle: Float[Array, ""]

    def __init__(
        self,
        offset_x_in_angstroms: FloatLike = 0.0,
        offset_y_in_angstroms: FloatLike = 0.0,
        phi_angle: FloatLike = 0.0,
        theta_angle: FloatLike = 0.0,
        psi_angle: FloatLike = 0.0,
        *,
        offset_z_in_angstroms: FloatLike | None = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `phi_angle`: Angle to rotate about first rotation axis, which is the z axis.
        - `theta_angle`: Angle to rotate about second rotation axis, which is the y axis.
        - `psi_angle`: Angle to rotate about third rotation axis, which is the z axis.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        if offset_z_in_angstroms is None:
            self.offset_in_angstroms = jnp.stack(
                (offset_x_in_angstroms, offset_y_in_angstroms), axis=-1, dtype=float
            )
        else:
            self.offset_in_angstroms = jnp.stack(
                (offset_x_in_angstroms, offset_y_in_angstroms, offset_z_in_angstroms),
                axis=-1,
                dtype=float,
            )
        self.phi_angle = jnp.asarray(phi_angle, dtype=float)
        self.theta_angle = jnp.asarray(theta_angle, dtype=float)
        self.psi_angle = jnp.asarray(psi_angle, dtype=float)

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate a `SO3` object from a set of Euler angles."""
        phi, theta, psi = self.phi_angle, self.theta_angle, self.psi_angle
        # Convert to radians.
        phi = jnp.deg2rad(phi)
        theta = jnp.deg2rad(theta)
        psi = jnp.deg2rad(psi)
        # Get sequence of rotations.
        R1, R2, R3 = (
            SO3.from_z_radians(phi),
            SO3.from_y_radians(theta),
            SO3.from_z_radians(psi),
        )
        return R3 @ R2 @ R1

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        phi_angle, theta_angle, psi_angle = convert_quaternion_to_euler_angles(
            rotation.wxyz, convention="zyz", extrinsic=True
        )
        return cls(phi_angle=phi_angle, theta_angle=theta_angle, psi_angle=psi_angle)

    @override
    def to_inverse_rotation(self) -> Self:
        return eqx.tree_at(
            lambda x: (x.phi_angle, x.theta_angle, x.psi_angle),
            self,
            (
                _negate_angle(self.psi_angle),
                _negate_angle(self.theta_angle),
                _negate_angle(self.phi_angle),
            ),
        )


class QuaternionPose(AbstractPose, strict=True):
    """An `AbstractPose` represented by unit quaternions."""

    offset_in_angstroms: Float[Array, "2"] | Float[Array, "3"]

    wxyz: Float[Array, "4"]

    def __init__(
        self,
        offset_x_in_angstroms: FloatLike = 0.0,
        offset_y_in_angstroms: FloatLike = 0.0,
        wxyz: Sequence[float] | Float[NDArrayLike, "4"] = (1.0, 0.0, 0.0, 0.0),
        *,
        offset_z_in_angstroms: FloatLike | None = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `wxyz`:
            The quaternion, represented as a vector $\\mathbf{q} = (q_w, q_x, q_y, q_z)$.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        if offset_z_in_angstroms is None:
            self.offset_in_angstroms = jnp.asarray(
                (offset_x_in_angstroms, offset_y_in_angstroms), dtype=float
            )
        else:
            self.offset_in_angstroms = jnp.asarray(
                (offset_x_in_angstroms, offset_y_in_angstroms, offset_z_in_angstroms),
                dtype=float,
            )
        if len(wxyz) != 4:
            raise ValueError(
                "Expected `wxyz` to be a sequence of floats with "
                f"length 4, but found `len(wxyz) = {len(wxyz)}`."
            )
        self.wxyz = jnp.asarray(wxyz, dtype=float)

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from the unit quaternion
        $\\mathbf{q} / |\\mathbf{q}|$.
        """
        # Generate SO3 object from unit quaternion
        R = SO3(wxyz=self.wxyz).normalize()
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        return cls(wxyz=rotation.wxyz)

    @override
    def to_inverse_rotation(self) -> Self:
        inverse_rotation = self.rotation.inverse()
        return eqx.tree_at(lambda _pose: _pose.wxyz, self, inverse_rotation.wxyz)


class AxisAnglePose(AbstractPose, strict=True):
    """An `AbstractPose` parameterized in the axis-angle representation.

    The axis-angle representation parameterizes elements of the so3 algebra,
    which are skew-symmetric matrices, with the euler vector
    $\\boldsymbol{\\omega} = (\\omega_x, \\omega_y, \\omega_z)$.
    The magnitude of this vector is the angle, and the unit vector is the axis.

    In a `SO3` object, the euler vector is mapped to SO3 group elements using
    the matrix exponential.
    """

    offset_in_angstroms: Float[Array, "2"] | Float[Array, "3"]

    euler_vector: Float[Array, "3"]

    def __init__(
        self,
        offset_x_in_angstroms: FloatLike = 0.0,
        offset_y_in_angstroms: FloatLike = 0.0,
        euler_vector: Sequence[float] | Float[NDArrayLike, "3"] = (0.0, 0.0, 0.0),
        *,
        offset_z_in_angstroms: FloatLike | None = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `euler_vector`:
            The axis-angle parameterization, represented with the euler
            vector $\\boldsymbol{\\omega}$.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        if offset_z_in_angstroms is None:
            self.offset_in_angstroms = jnp.asarray(
                (offset_x_in_angstroms, offset_y_in_angstroms), dtype=float
            )
        else:
            self.offset_in_angstroms = jnp.asarray(
                (offset_x_in_angstroms, offset_y_in_angstroms, offset_z_in_angstroms),
                dtype=float,
            )
        if len(euler_vector) != 3:
            raise ValueError(
                "Expected `euler_vector` to be a sequence of floats with "
                f"length 3, but found `len(euler_vector) = {len(euler_vector)}`."
            )
        self.euler_vector = jnp.asarray(euler_vector, dtype=float)

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from an euler vector using the exponential map."""
        # Convert degrees to radians
        euler_vector = jnp.deg2rad(self.euler_vector)
        # Project the tangent vector onto the manifold with
        # the exponential map
        R = SO3.exp(euler_vector)
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        # Compute the euler vector from the logarithmic map
        euler_vector = jnp.rad2deg(rotation.log())
        return cls(euler_vector=euler_vector)

    @override
    def to_inverse_rotation(self) -> Self:
        return self.from_rotation_and_translation(
            self.rotation.inverse(), self.offset_in_angstroms
        )


def _negate_angle(angle):
    return ((-angle + 180) % 360) - 180



from ._geometry import cartesian_to_polar as cartesian_to_polar
from ._make_coordinate_grids import (
    make_1d_coordinate_grid as make_1d_coordinate_grid,
    make_1d_frequency_grid as make_1d_frequency_grid,
    make_coordinate_grid as make_coordinate_grid,
    make_frequency_grid as make_frequency_grid,
    make_frequency_slice as make_frequency_slice,
    make_radial_coordinate_grid as make_radial_coordinate_grid,
    make_radial_frequency_grid as make_radial_frequency_grid,
)



"""
Routines for starfile serialization and deserialization.
"""

import pathlib
from typing import Any, Literal, cast

import pandas as pd
import starfile


def read_starfile(filename: str | pathlib.Path, **kwargs: Any) -> dict[str, pd.DataFrame]:
    """Read a STAR file, safely."""
    # Make sure filename is valid starfile
    _validate_filename(filename, mode="r")
    # Read starfile
    path_to_filename = pathlib.Path(filename)
    starfile_data = starfile.read(path_to_filename, always_dict=True, **kwargs)
    return cast(dict[str, pd.DataFrame], starfile_data)


def write_starfile(starfile_data, filename: str | pathlib.Path, **kwargs: Any):
    """Write a STAR file, safely."""
    # Make sure filename is valid starfile
    _validate_filename(filename, mode="w")
    # Write starfile
    path_to_filename = pathlib.Path(filename)
    return starfile.write(starfile_data, path_to_filename, **kwargs)  # type: ignore


def _validate_filename(filename: str | pathlib.Path, mode: Literal["r", "w"]):
    suffixes = pathlib.Path(filename).suffixes
    if not (len(suffixes) == 1 and suffixes[0] == ".star"):
        raise OSError(
            f"Tried to {('write' if mode == 'w' else 'read')} STAR file, "
            "but the filename does not include a '.star' "
            f"suffix. Got filename '{filename}'."
        )



from time import time

import cryojax.ndimage as im
import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cryojax.constants import PengScatteringFactorParameters
from cryojax.io import read_atoms_from_pdb


def setup_volumes_and_configs(n_iterations, n_atoms, box_size, pixel_size=2.0):
    """Setup volumes and image configs for different test conditions."""
    # Read atoms (subsample to get desired n_atoms). Make sure to pull from
    # git LFS
    atom_positions, atom_types, atom_properties = read_atoms_from_pdb(
        "../docs/examples/data/thyroglobulin_initial.pdb",
        center=True,
        loads_properties=True,
        selection_string="name CA",
    )

    # Subsample atoms if needed
    if n_atoms < len(atom_positions):
        indices = jax.random.choice(
            jax.random.key(42), len(atom_positions), (n_atoms,), replace=False
        )
        atom_positions = atom_positions[indices]
        atom_types = np.asarray([atom_types[i] for i in indices])
        atom_properties = {k: v[indices] for k, v in atom_properties.items()}  # type: ignore

    # Create volumes
    scattering_parameters = PengScatteringFactorParameters(atom_types)
    volume_gmm = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        scattering_parameters,
        extra_b_factors=atom_properties["b_factors"],
    )

    # Fourier slice volume (pre-computed grid)
    real_voxel_grid = volume_gmm.to_real_voxel_grid(
        shape=(box_size, box_size, box_size), voxel_size=pixel_size
    )

    times = []
    for _ in range(n_iterations + 1):
        start_time = time()
        # The volume rendering is already done in setup, but we simulate the cost
        _ = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid, pad_scale=2)
        jax.block_until_ready(_)
        times.append(time() - start_time)
    avg_time = np.mean(times[1:])
    print(f"Average time to create FourierVoxelGridVolume: {avg_time * 1000:.2f} ms")

    volume_fourier_grid = cxs.FourierVoxelGridVolume.from_real_voxel_grid(
        real_voxel_grid, pad_scale=2
    )

    # Atom volume for direct projection
    atom_volume = cxs.IndependentAtomVolume(
        positions=atom_positions,
        scattering_factors=im.FourierGaussian(amplitude=1.0, b_factor=10.0),
    )

    # Image config
    pad_options = dict(shape=(int(box_size * 1.5), int(box_size * 1.5)))
    config = cxs.BasicImageConfig(
        shape=(box_size, box_size),
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
        pad_options=pad_options,
    )

    return volume_fourier_grid, avg_time, volume_gmm, atom_volume, config


@eqx.filter_jit
@eqx.filter_vmap(in_axes=(eqx.if_array(0), eqx.if_array(0), eqx.if_array(0), None, None))
def simulate_image_batch(image_config, pose, transfer_theory, volume, integrator):
    """Simulate a batch of images."""
    image_model = cxs.make_image_model(
        volume=volume,
        image_config=image_config,
        pose=pose,
        transfer_theory=transfer_theory,
        volume_integrator=integrator,
    )
    return image_model.simulate()


def generate_random_parameters(n_projections, config):
    """Generate random poses and CTF parameters."""
    from cryojax.rotations import SO3

    keys = jax.random.split(jax.random.key(12345), n_projections)

    def make_single_params(key):
        key, subkey = jax.random.split(key)
        rotation = SO3.sample_uniform(subkey)

        key, subkey = jax.random.split(key)
        ny, nx = config.shape
        offset_in_angstroms = (
            jax.random.uniform(subkey, (2,), minval=-0.1, maxval=0.1)
            * jnp.asarray((nx, ny))
            / 2
            * config.pixel_size
        )

        pose = cxs.EulerAnglePose.from_rotation_and_translation(
            rotation, offset_in_angstroms
        )

        key, subkey = jax.random.split(key)
        defocus = jax.random.uniform(subkey, (), minval=10000, maxval=15000)

        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(
                defocus_in_angstroms=defocus,
                astigmatism_in_angstroms=50.0,
                astigmatism_angle=0.0,
                spherical_aberration_in_mm=2.7,
            ),
            amplitude_contrast_ratio=0.1,
        )

        return config, pose, transfer_theory

    configs, poses, transfer_theories = eqx.filter_vmap(make_single_params)(keys)
    return configs, poses, transfer_theories


def benchmark_projection_methods(
    n_projections_list, n_atoms_list, box_sizes, n_iterations
):
    """Benchmark both projection methods across different conditions."""
    results = []

    for n_atoms in n_atoms_list:
        for box_size in box_sizes:
            print(f"Testing n_atoms={n_atoms}, box_size={box_size}")

            # Setup volumes
            volume_fourier_grid, volume_render_time, gmm_volume, atom_volume, config = (
                setup_volumes_and_configs(n_iterations, n_atoms, box_size)
            )

            for n_projections in n_projections_list:
                print(f"  n_projections={n_projections}")

                # Generate parameters
                configs, poses, transfer_theories = generate_random_parameters(
                    n_projections, config
                )

                # Benchmark Fourier Slice Extraction
                times = []
                integrator = cxs.FourierSliceExtraction()
                for _ in range(n_iterations + 1):
                    start_time = time()
                    images = simulate_image_batch(
                        configs,
                        poses,
                        transfer_theories,
                        volume_fourier_grid,
                        integrator,
                    )
                    images.block_until_ready()
                    times.append(time() - start_time)

                fs_time_per_projection = np.mean(times[1:])
                fs_total_time = volume_render_time + fs_time_per_projection

                # Benchmark Atom Projection (FFT)
                times = []
                integrator = cxs.FFTAtomProjection(eps=1e-16)
                for _ in range(n_iterations + 1):
                    start_time = time()
                    images = simulate_image_batch(
                        configs,
                        poses,
                        transfer_theories,
                        atom_volume,
                        integrator,
                    )
                    images.block_until_ready()
                    times.append(time() - start_time)

                atom_time_total = np.mean(times[1:])

                # Benchmark GMM Projection (FFT)
                times = []
                integrator = cxs.GaussianMixtureProjection()
                for _ in range(n_iterations + 1):
                    start_time = time()
                    images = simulate_image_batch(
                        configs,
                        poses,
                        transfer_theories,
                        gmm_volume,
                        integrator,
                    )
                    images.block_until_ready()
                    times.append(time() - start_time)

                gmm_time_total = np.mean(times[1:])

                results.append(
                    {
                        "n_atoms": n_atoms,
                        "box_size": box_size,
                        "n_projections": n_projections,
                        "fourier_slice_time": fs_total_time,
                        "atom_projection_time": atom_time_total,
                        "fs_volume_render_time": volume_render_time,
                        "fs_projection_time": fs_time_per_projection,
                        "gmm_projection_time": gmm_time_total,
                    }
                )

    return pd.DataFrame(results)


def plot_crossover_analysis(df, datetimestamp):
    """Plot the crossover analysis showing when each method is faster."""
    n_atoms_list = df["n_atoms"].unique()
    box_sizes = df["box_size"].unique()

    fig, axes = plt.subplots(
        len(n_atoms_list),
        len(box_sizes),
        figsize=(4 * len(box_sizes), 4 * len(n_atoms_list)),
        squeeze=False,
    )

    for i, n_atoms in enumerate(n_atoms_list):
        for j, box_size in enumerate(box_sizes):
            ax = axes[i, j]

            subset = df[(df["n_atoms"] == n_atoms) & (df["box_size"] == box_size)]

            ax.plot(
                subset["n_projections"],
                subset["fourier_slice_time"] * 1000,
                "o-",
                label="Fourier Slice",
                color="blue",
                linewidth=2,
            )
            ax.plot(
                subset["n_projections"],
                subset["atom_projection_time"] * 1000,
                "s-",
                label="Atom Projection",
                color="red",
                linewidth=2,
            )
            ax.plot(
                subset["n_projections"],
                subset["gmm_projection_time"] * 1000,
                "^-",
                label="GMM Projection",
                color="green",
                linewidth=2,
            )

            ax.set_xlabel("Number of Projections")
            ax.set_ylabel("Total Time (ms)")
            ax.set_title(f"{n_atoms} atoms, {box_size}×{box_size} box")
            ax.legend()
            ax.grid(True, alpha=0.3)
            ax.set_xscale("log")
            ax.set_yscale("log")

    plt.tight_layout()
    plt.savefig(
        f"benchmark_projection_method_tradeoff_{datetimestamp}.png",
        dpi=300,
        bbox_inches="tight",
    )
    plt.show()


def find_crossover_points(df):
    """Find crossover points where methods have equal performance."""
    crossovers = []

    for (n_atoms, box_size), group in df.groupby(["n_atoms", "box_size"]):
        group = group.sort_values("n_projections")

        fs_times = group["fourier_slice_time"].values
        gmm_times = group["gmm_projection_time"].values
        n_proj = group["n_projections"].values

        # Find where lines cross
        diff = fs_times - gmm_times
        sign_changes = np.where(np.diff(np.signbit(diff)))[0]

        for idx in sign_changes:
            # Linear interpolation to find precise crossover
            x1, x2 = n_proj[idx], n_proj[idx + 1]
            y1, y2 = diff[idx], diff[idx + 1]
            crossover_n_proj = x1 - y1 * (x2 - x1) / (y2 - y1)

            crossovers.append(
                {
                    "n_atoms": n_atoms,
                    "box_size": box_size,
                    "crossover_n_projections": crossover_n_proj,
                }
            )

    return pd.DataFrame(crossovers)


if __name__ == "__main__":
    # Test parameters
    n_projections_list = [1, 3, 10, 30, 100]
    n_atoms_list = [30, 100, 300]
    box_sizes = [32, 64]

    print("Running projection method crossover benchmark...")
    print("This will test Fourier slicing vs GMM projection across different conditions")

    # Run benchmark
    results_df = benchmark_projection_methods(
        n_projections_list, n_atoms_list, box_sizes, n_iterations=3
    )

    # Save results
    datetimestamp = pd.Timestamp.now().strftime("%Y%m%d_%H%M%S")
    results_df.to_csv(
        f"benchmark_projection_method_tradeoff_{datetimestamp}.csv", index=False
    )
    print(f"Results saved to benchmark_projection_method_tradeoff_{datetimestamp}.csv")

    # Plot results
    plot_crossover_analysis(results_df, datetimestamp)

    # Find and display crossover points
    crossovers = find_crossover_points(results_df)
    print("\nCrossover points (where methods have equal performance):")
    print(crossovers)

    # Summary statistics
    print("\nSummary:")
    for (n_atoms, box_size), group in results_df.groupby(["n_atoms", "box_size"]):
        fs_faster = (group["fourier_slice_time"] < group["gmm_projection_time"]).sum()
        gmm_faster = len(group) - fs_faster
        print(
            f"{n_atoms} atoms, {box_size}×{box_size}: "
            f"Fourier slice faster in {fs_faster}/{len(group)} cases, "
            f"GMM projection faster in {gmm_faster}/{len(group)} cases"
        )



from abc import abstractmethod

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ..._internal import error_if_negative
from ...constants import wavelength_from_kilovolts
from ...jax_util import FloatLike
from .common_functions import (
    compute_phase_shift_from_amplitude_contrast_ratio,
    compute_phase_shifts_with_spherical_aberration,
)


class AbstractCTF(eqx.Module, strict=True):
    """An abstract base class for a CTF in cryo-EM."""

    @abstractmethod
    def compute_aberration_phase_shifts(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        wavelength_in_angstroms: FloatLike,
        defocus_offset: FloatLike | None = None,
    ) -> Float[Array, "y_dim x_dim"]:
        raise NotImplementedError

    def __call__(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        voltage_in_kilovolts: FloatLike,
        amplitude_contrast_ratio: FloatLike = 0.1,
        phase_shift: FloatLike = 0.0,
        outputs_exp: bool = False,
        defocus_offset: FloatLike | None = None,
    ) -> Float[Array, "y_dim x_dim"] | Complex[Array, "y_dim x_dim"]:
        """Compute the CTF as a JAX array.

        **Arguments:**

        - `frequency_grid_in_angstroms`:
            The grid of frequencies in units of inverse angstroms. This can
            be computed with [`cryojax.coordinates.make_frequency_grid`](https://mjo22.github.io/cryojax/api/coordinates/making_coordinates/#cryojax.coordinates.make_frequency_grid)
        - `voltage_in_kilovolts`:
            The accelerating voltage of the microscope in kilovolts. This
            is converted to the wavelength of incident electrons using
            the function [`cryojax.constants.convert_keV_to_angstroms`](https://mjo22.github.io/cryojax/api/constants/units/#cryojax.constants.convert_keV_to_angstroms)
        - `amplitude_contrast_ratio`:
            The amplitude contrast ratio. This argument is not used if `outputs_exp = True`, as
            the amplitude contrast ratio cannot simply be absorbed into a phase shift.
        - `phase_shift`:
            Additional constant phase shift applied to the frequency-dependent phase shifts.
        - `outputs_exp`:
            If `False`, return the CTF used in linear image formation theory. If `True`, return
            the CTF (or wave transfer function) as a complex exponential.
        """  # noqa: E501
        # Get the wavelength
        wavelength_in_angstroms = wavelength_from_kilovolts(
            jnp.asarray(voltage_in_kilovolts)
        )
        # Frequency-dependent phase shifts
        aberration_phase_shifts = self.compute_aberration_phase_shifts(
            frequency_grid_in_angstroms,
            wavelength_in_angstroms=wavelength_in_angstroms,
            defocus_offset=defocus_offset,
        )
        # Constant phase shift, convert degrees to radians
        phase_shift = jnp.deg2rad(phase_shift)
        if not outputs_exp:
            # Compute the CTF
            amplitude_contrast_phase_shift = (
                compute_phase_shift_from_amplitude_contrast_ratio(
                    jnp.asarray(amplitude_contrast_ratio, dtype=float)
                )
            )
            return jnp.sin(
                aberration_phase_shifts - (phase_shift + amplitude_contrast_phase_shift)
            )
        else:
            # Compute the "complex CTF", correcting for the amplitude contrast
            # and additional phase shift in the zero mode
            return jnp.exp(-1.0j * (aberration_phase_shifts - phase_shift))


class AstigmaticCTF(AbstractCTF, strict=True):
    """Compute an astigmatic Contrast Transfer Function (CTF) with a
    spherical aberration correction and amplitude contrast ratio.

    !!! info
        `cryojax` uses a convention different from CTFFIND for
        astigmatism parameters. CTFFIND returns defocus major and minor
        axes, called "defocus1" and "defocus2". In order to convert
        from CTFFIND to `cryojax`,

        ```python
        defocus1, defocus2 = ... # Read from CTFFIND
        ctf = AstigmaticCTF(
            defocus_in_angstroms=(defocus1+defocus2)/2,
            astigmatism_in_angstroms=defocus1-defocus2,
            ...
        )
        ```
    """

    defocus_in_angstroms: Float[Array, ""]
    astigmatism_in_angstroms: Float[Array, ""]
    astigmatism_angle: Float[Array, ""]
    spherical_aberration_in_mm: Float[Array, ""]

    def __init__(
        self,
        defocus_in_angstroms: FloatLike = 10000.0,
        astigmatism_in_angstroms: FloatLike = 0.0,
        astigmatism_angle: FloatLike = 0.0,
        spherical_aberration_in_mm: FloatLike = 2.7,
    ):
        """**Arguments:**

        - `defocus_in_angstroms`: The mean defocus in Angstroms.
        - `astigmatism_in_angstroms`: The amount of astigmatism in Angstroms.
        - `astigmatism_angle`: The defocus angle.
        - `spherical_aberration_in_mm`: The spherical aberration coefficient in mm.
        """
        self.defocus_in_angstroms = jnp.asarray(defocus_in_angstroms, dtype=float)
        self.astigmatism_in_angstroms = jnp.asarray(astigmatism_in_angstroms, dtype=float)
        self.astigmatism_angle = jnp.asarray(astigmatism_angle, dtype=float)
        self.spherical_aberration_in_mm = jnp.asarray(
            spherical_aberration_in_mm, dtype=float
        )

    def compute_aberration_phase_shifts(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        wavelength_in_angstroms: FloatLike,
        defocus_offset: FloatLike | None = None,
    ) -> Float[Array, "y_dim x_dim"]:
        """Compute the frequency-dependent phase shifts due to wave aberration.

        This is often denoted as $\\chi(\\boldsymbol{q})$ for the in-plane
        spatial frequency $\\boldsymbol{q}$.

        **Arguments:**

        - `frequency_grid_in_angstroms`:
            The grid of frequencies in units of inverse angstroms. This can
            be computed with [`cryojax.coordinates.make_frequency_grid`](https://mjo22.github.io/cryojax/api/coordinates/making_coordinates/#cryojax.coordinates.make_frequency_grid)
        - `voltage_in_kilovolts`:
            The accelerating voltage of the microscope in kilovolts. This
            is converted to the wavelength of incident electrons using
            the function [`cryojax.constants.convert_keV_to_angstroms`](https://mjo22.github.io/cryojax/api/constants/units/#cryojax.constants.convert_keV_to_angstroms)
        - `defocus_offset`:
            An optional defocus offset to apply to the `defocus_in_angstroms` at runtime.
        """
        astigmatism_angle = jnp.deg2rad(self.astigmatism_angle)
        # Convert spherical abberation coefficient to angstroms
        spherical_aberration_in_angstroms = (
            error_if_negative(self.spherical_aberration_in_mm) * 1e7
        )
        # Compute phase shifts for CTF
        phase_shifts = compute_phase_shifts_with_spherical_aberration(
            frequency_grid_in_angstroms,
            (
                self.defocus_in_angstroms
                if defocus_offset is None
                else self.defocus_in_angstroms + jnp.asarray(defocus_offset, dtype=float)
            ),
            self.astigmatism_in_angstroms,
            astigmatism_angle,
            jnp.asarray(wavelength_in_angstroms, dtype=float),
            spherical_aberration_in_angstroms,
        )
        return phase_shifts



"""
Abstraction of a rotation.
"""

from abc import abstractmethod
from typing import overload
from typing_extensions import Self

from equinox import AbstractClassVar, Module
from jaxtyping import Array, PRNGKeyArray


class AbstractRotation(Module, strict=True):
    """Base class for a rotation."""

    space_dimension: AbstractClassVar[int]

    @overload
    def __matmul__(self, other: Self) -> Self: ...

    @overload
    def __matmul__(self, other: Array) -> Array: ...

    def __matmul__(self, other: Self | Array) -> Self | Array:
        """The `@` operator.

        Can either compose with other `AbstractRotation`s or
        act on arrays.
        """
        if isinstance(other, Array):
            return self.apply(other)
        elif isinstance(other, AbstractRotation):
            if not self.space_dimension == other.space_dimension:
                raise ValueError(
                    f"Rotation of type {type(other).__name__} is "
                    f"incompatible with type {type(self).__name__}. "
                    "Tried to compose these rotations with A @ B syntax."
                )
            return self.compose(other)
        else:
            raise ValueError(
                "Allowed types for `@` operator are JAX arrays and other "
                f"rotations. Tried to compose with type {type(other).__name__}."
            )

    @abstractmethod
    def apply(self, target: Array) -> Array:
        """Apply the rotation to an array."""
        raise NotImplementedError

    @abstractmethod
    def compose(self, other) -> Self:
        """Composes this transformation with another."""
        raise NotImplementedError

    @abstractmethod
    def inverse(self) -> Self:
        """Get the inverse of this rotation."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def identity(cls: type[Self]) -> Self:
        """Return the identity element."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def sample_uniform(cls: type[Self], key: PRNGKeyArray) -> Self:
        """Draw a uniform sample."""
        raise NotImplementedError
